"""A collection of different strategies for iterating over streams
of :class:`~.Scan`-like objects.
"""
import bisect
import warnings

from typing import Any, Callable, DefaultDict, Deque, Generic, Iterator, List, Optional, TypeVar, Union

from collections import deque, defaultdict
from six import PY2

from . import ScanBunch


def _null_scan_validator(scan):
    return True


def _null_scan_cacher(scan):
    return None


ITERATION_MODE_GROUPED = "grouped"
ITERATION_MODE_SINGLE = "single"


ScanGroupType = TypeVar("ScanGroupType")
ScanType = TypeVar("ScanType")


class _ScanIteratorImplBase(Generic[ScanType, ScanGroupType]):
    """Internal base class for scan iteration strategies

    Attributes
    ----------
    iterator : :class:`Iterator`
        An iterator that produces a raw scan which will be packaged by :attr:`scan_packer`
    scan_cacher : :class:`Callable`
        A callable that will cache the scan, or a no-op
    scan_packer : :class:`Callable`
        A callable that will package a raw scan object into a :class:`Scan` object
    scan_validator : :class:`Callable`
        A callable that will check if a raw scan object is valid or not for filtering
        out unwanted entries, or a no-op
    """

    iterator: Iterator[Any]
    _producer: Union[Iterator[ScanType], Iterator[ScanGroupType]]

    def __init__(self, iterator, scan_packer, scan_validator=None, scan_cacher=None, **kwargs):
        if scan_validator is None:
            scan_validator = _null_scan_validator
        if scan_cacher is None:
            scan_cacher = _null_scan_cacher
        self.iterator = iterator
        self.scan_packer = scan_packer
        self.scan_validator = scan_validator
        self.scan_cacher = scan_cacher
        self._producer = None

    def __iter__(self):
        return self

    def next(self):
        """Py2 compatible iterator
        """
        if self._producer is None:
            self._producer = self._make_producer()
        return next(self._producer)

    def __next__(self):
        return self.next() # pylint: disable=not-callable

    def _make_producer(self) -> Union[Iterator[ScanType], Iterator[ScanGroupType]]:
        raise NotImplementedError()

    @property
    def iteration_mode(self):
        """The iteration mode of the strategy, "grouped" or "single"
        """
        raise NotImplementedError()

    @classmethod
    def from_scan_source(cls, iterator, scan_source, **kwargs):
        """Create an iterator strategy for `iterator` from `scan_source`

        Parameters
        ----------
        iterator : :class:`Iterable`
            The iterator over raw scan data to consume
        scan_source : :class:`~.ScanIterator`
            The data extraction wrapper to provide with an iteration strategy
        """
        return cls(iterator, scan_source._make_scan, scan_source._validate, scan_source._cache_scan, **kwargs)


class _SingleScanIteratorImpl(_ScanIteratorImplBase[ScanType, ScanGroupType], Iterator[ScanType]):
    """Iterate over individual scans.

    The default strategy when MS1 scans are missing.
    """

    _producer: Iterator[ScanType]

    @property
    def iteration_mode(self):
        return ITERATION_MODE_SINGLE

    def _make_producer(self) -> Iterator[ScanType]:
        _make_scan = self.scan_packer
        _validate = self.scan_validator
        _cache_scan = self.scan_cacher
        for scan in self.iterator:
            packed = _make_scan(scan)
            if not _validate(packed):
                continue
            _cache_scan(packed)
            yield packed


class _FakeGroupedScanIteratorImpl(_SingleScanIteratorImpl[ScanType, ScanGroupType], Iterator[ScanGroupType]):
    """Mimics the interface of :class:`_GroupedScanIteratorImpl` for
    scan sequences which only support single scans, or which do not
    guarantee sequential access to precursor/product collections.
    """

    _producer: Iterator[ScanGroupType]

    @property
    def iteration_mode(self):
        return ITERATION_MODE_GROUPED

    def _make_producer(self) -> Iterator[ScanGroupType]:
        generator = super(_FakeGroupedScanIteratorImpl, self)._make_producer()
        for scan in generator:
            if scan.ms_level == 1:
                yield ScanBunch(scan, [])
            else:
                yield ScanBunch(None, [scan])


class _GroupedScanIteratorImpl(_ScanIteratorImplBase[ScanType, ScanGroupType], Iterator[ScanGroupType]):
    """Iterate over related scan bunches.

    The default strategy when MS1 scans are known to be
    present, even if MSn scans are not.
    """

    _producer: Iterator[ScanGroupType]

    @property
    def iteration_mode(self):
        return ITERATION_MODE_GROUPED

    def _make_producer(self) -> Iterator[ScanGroupType]:
        _make_scan = self.scan_packer
        _validate = self.scan_validator
        _cache_scan = self.scan_cacher

        precursor_scan = None
        product_scans = []

        current_level = 1

        for scan in self.iterator:
            packed = _make_scan(scan)
            if not _validate(packed):
                continue
            _cache_scan(packed)
            if packed.ms_level > 1:
                # inceasing ms level
                if current_level < packed.ms_level:
                    current_level = packed.ms_level
                # decreasing ms level
                elif current_level > packed.ms_level:
                    current_level = packed.ms_level
                product_scans.append(packed)
            elif packed.ms_level == 1:
                if current_level > 1:
                    if precursor_scan is not None:
                        precursor_scan.product_scans = list(product_scans)
                    yield ScanBunch(precursor_scan, product_scans)
                else:
                    if precursor_scan is not None:
                        precursor_scan.product_scans = list(product_scans)
                        yield ScanBunch(precursor_scan, product_scans)
                precursor_scan = packed
                product_scans = []
            else:
                raise ValueError("Could not interpret MS Level %r" %
                                 (packed.ms_level,))
        if precursor_scan is not None:
            yield ScanBunch(precursor_scan, product_scans)


class GenerationTracker(object):
    _generations_type = list if PY2 else deque

    def __init__(self):
        self.generation_to_id = defaultdict(set)
        self.id_to_generation = dict()
        self.generations = self._generations_type()

    def _add_generation(self, generation):
        bisect.insort_left(self.generations, generation)

    def clear(self):
        self.generation_to_id.clear()
        self.id_to_generation.clear()
        self.generations = self._generations_type()

    def add(self, identifier, generation):
        if generation not in self.generation_to_id:
            self._add_generation(generation)
        self.generation_to_id[generation].add(identifier)
        self.id_to_generation[identifier] = generation

    def remove(self, identifier):
        if identifier not in self.id_to_generation:
            return False
        generation = self.id_to_generation[identifier]
        self.generation_to_id[generation].remove(identifier)
        del self.id_to_generation[identifier]
        if len(self.generation_to_id[generation]) == 0:
            self.generations.remove(generation)
        return True

    def older_than(self, generation):
        result = []
        purged = []
        for gen in self.generations:
            if gen < generation:
                members = self.generation_to_id[gen]
                result.extend(members)
                purged.append(gen)
            else:
                break
        for r in result:
            del self.id_to_generation[r]
        for gen in purged:
            del self.generation_to_id[gen]
            self.generations.remove(gen)
        return result


class _InterleavedGroupedScanIteratorImpl(_GroupedScanIteratorImpl[ScanType, ScanGroupType]):
    """Iterate over related scan bunches.

    The default strategy when MS1 scans are known to be
    present, even if MSn scans are not.
    """

    buffering: int
    ms1_buffer: Deque[ScanType]
    product_mapping: DefaultDict[str, List[ScanType]]

    generation_tracker: GenerationTracker
    orphans: List[ScanType]

    passed_first_ms1: bool
    highest_ms_level: int
    generation: int

    def __init__(self, iterator, scan_packer, scan_validator=None, scan_cacher=None, buffering=5):
        super(_InterleavedGroupedScanIteratorImpl, self).__init__(
            iterator, scan_packer, scan_validator, scan_cacher)
        if buffering < 2:
            raise ValueError("Interleaved buffering must be greater than 1")
        self.buffering = buffering
        self.ms1_buffer = deque()
        self.product_mapping = defaultdict(list)
        self.generation_tracker = GenerationTracker()
        self.orphans = []
        self.passed_first_ms1 = False
        self.highest_ms_level = 0
        self.generation = 0

    def pop_precursor(self, precursor_id: str) -> List[ScanType]:
        self.generation_tracker.remove(precursor_id)
        return self.product_mapping.pop(precursor_id, [])

    def deque_group(self, flush_products=False) -> ScanGroupType:
        """Remove the next scan from the MS1 queue, grouped with
        any associated MSn scans.

        Parameters
        ----------
        flush_products : bool
            Whether to flush all the remaining product scans with this
            group.

        Returns
        -------
        ScanBunch
        """
        precursor = self.ms1_buffer.popleft()
        _empty = []
        products = self.pop_precursor(precursor.id)
        if None in self.product_mapping:
            products += self.product_mapping.pop(None, _empty)

        # Flush out older precursors' products that haven't turned up yet, they
        # probably aren't coming soon.
        for prec_id in self.generation_tracker.older_than(self.generation - self.buffering):
            products.extend(self.product_mapping.pop(prec_id, _empty))

        # Look for MSn for n > 2
        if self.highest_ms_level > 2:
            extra_blocks = [products]
            for _ in range(self.highest_ms_level - 2):
                new_block = []
                for prod in extra_blocks[-1]:
                    new_block.extend(self.pop_precursor(prod.id))
                if not new_block:
                    break
                extra_blocks.append(new_block)
            if len(extra_blocks) > 1:
                products = extra_blocks[0]
                for block in extra_blocks[1:]:
                    products.extend(block)

        if flush_products:
            if self.product_mapping:
                lingering = {s.id for ss in self.product_mapping.values() for s in ss}
                missing = set(self.product_mapping)
                unclaimed_precursors = missing - lingering
                warnings.warn("Lingering Product Sets For %r!" %
                              (sorted(unclaimed_precursors), ))
            for _, value in self.product_mapping.items():
                products += value
            self.product_mapping.clear()
            self.generation_tracker.clear()
        precursor.product_scans = products

        # Collect any MSn spectra which pre-date the first precursor if they are encountered before
        # the first precursor is found.
        if not self.passed_first_ms1 and self.ms1_buffer:
            current_ms1_time = precursor.scan_time
            for prec_id, prods in list(self.product_mapping.items()):
                masked = set()
                for i, prod in enumerate(prods):
                    if prod.scan_time <= current_ms1_time:
                        products.append(prod)
                        masked.add(i)
                # We've only removed some of the products under this precursor, so just
                # remove those products from mapping.
                if len(masked) < len(prods):
                    prods = [v for i, v in enumerate(prods) if i not in masked]
                    self.product_mapping[prec_id] = prods
                else:
                    # Otherwise we must have completely consumed the products of this
                    # precursor, so we need to remove it from the tracking.
                    self.pop_precursor(prec_id)
            self.passed_first_ms1 = True


        self.generation += 1
        return ScanBunch(precursor, products)

    def add_product(self, scan: ScanType):
        """Add MSn scan to :attr:`product_mapping` for the associated
        precursor scan ID.

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The scan to track.
        """
        pinfo = scan.precursor_information
        if pinfo is None:
            precursor_id = None
        else:
            precursor_id = pinfo.precursor_scan_id
        if precursor_id is None:
            try:
                precursor_id = self.ms1_buffer[-1].id
            except IndexError:
                precursor_id = None
        if precursor_id not in self.product_mapping:
            self.generation_tracker.add(precursor_id, self.generation)
        self.product_mapping[precursor_id].append(scan)

    def add_precursor(self, scan: ScanType) -> bool:
        """Add MS1 scan to :attr:`ms1_buffer`

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The scan to track.

        Returns
        -------
        buffer_full : bool
            Whether or not :attr:`ms1_buffer` is full.
        """
        self.ms1_buffer.append(scan)
        return len(self.ms1_buffer) >= self.buffering

    def _make_producer(self) -> Iterator[ScanGroupType]:
        _make_scan = self.scan_packer
        _validate = self.scan_validator
        _cache_scan = self.scan_cacher

        current_level = 1

        for scan in self.iterator:
            packed = _make_scan(scan)
            if not _validate(packed):
                continue
            _cache_scan(packed)
            if packed.ms_level > 1:
                # inceasing ms level
                if current_level < packed.ms_level:
                    current_level = packed.ms_level
                    if current_level > self.highest_ms_level:
                        self.highest_ms_level = current_level

                # decreasing ms level
                elif current_level > packed.ms_level:
                    current_level = packed.ms_level
                self.add_product(packed)
            elif packed.ms_level == 1:
                do_emit = self.add_precursor(packed)
                if do_emit:
                    yield self.deque_group()
            else:
                raise ValueError("Could not interpret MS Level %r" %
                                 (packed.ms_level,))

        while len(self.ms1_buffer) > 1:
            yield self.deque_group()

        if self.ms1_buffer:
            yield self.deque_group(flush_products=True)


class MSEIterator(_GroupedScanIteratorImpl[ScanType, ScanGroupType]):
    """A scan iterator implementation for grouping MS^E spectra according
    to the specified functions.

    Attributes
    ----------
    low_energy_config : int
        The function corresponding to lower energy. These correspond
        to the MS1 equivalent.
    lock_mass_config : int
        The function corresponding to the lockmass. Lockmass scans
        will be skipped.
    """

    low_energy_config: int = 1
    lock_mass_config: int = 3
    low_energy_config: Optional[Callable]

    def __init__(self, iterator, scan_packer, scan_validator=None,
                 scan_cacher=None, low_energy_config=1, lock_mass_config=3,
                 on_lock_mass_scan: Optional[Callable] = None, **kwargs):
        super(MSEIterator, self).__init__(
            iterator, scan_packer, scan_validator, scan_cacher)
        self.low_energy_config = low_energy_config
        self.lock_mass_config = lock_mass_config
        self.on_lock_mass_scan = on_lock_mass_scan

    def _make_producer(self) -> Iterator[ScanGroupType]:
        _make_scan = self.scan_packer
        _validate = self.scan_validator
        _cache_scan = self.scan_cacher

        precursor_scan = None
        product_scans = []

        current_level = 'low'

        for scan in self.iterator:
            packed = _make_scan(scan)
            if not _validate(packed):
                continue
            _cache_scan(packed)
            config = packed.acquisition_information[0].scan_configuration
            if config == self.lock_mass_config:
                if self.on_lock_mass_scan is not None:
                    self.on_lock_mass_scan(scan)
                continue

            if config != self.low_energy_config:
                if current_level == 'low':
                    current_level = 'high'
                # decreasing ms level
                product_scans.append(packed)
            elif config == self.low_energy_config:
                if current_level != 'low':
                    yield ScanBunch(precursor_scan, product_scans)
                else:
                    if precursor_scan is not None:
                        yield ScanBunch(precursor_scan, product_scans)
                precursor_scan = packed
                product_scans = []
            else:
                raise ValueError("Could not interpret MS Level %r" %
                                 (packed.ms_level,))
        if precursor_scan is not None:
            yield ScanBunch(precursor_scan, product_scans)



# -*- coding: utf-8 -*-

import math
import numpy as np
import operator

from .utils import Base

eps = 1e-4


class IsotopicFitRecord(object):
    """Describes a single isotopic pattern fit, comparing how well an
    experimentally observed sequence of peaks matches a theoretical isotopic
    pattern.

    IsotopicFitRecord instances are hashable and orderable (by score).

    Attributes
    ----------
    charge : int
        The charge state used to generate the theoretical pattern
    data : object
        An arbitrary Python object containing extra information
    experimental : list of FittedPeak
        The observed experimental peaks to be fitted
    missed_peaks : int
        The number of peaks in the theoretical pattern that do not
        have a matching experimental peak
    monoisotopic_peak : FittedPeak
        The fitted peak which corresponds to the monoisotopic peak
    score : float
        The score assigned to the fit by an IsotopicFitter object
    seed_peak : FittedPeak
        The peak that was used to initiate the fit. This may be unused
        if not using an Averagine method
    theoretical : list of TheoreticalPeak
        The theoretical isotopic pattern being fitted on the experimental data
    """

    __slots__ = ["seed_peak", "score", "charge", "experimental", "theoretical",
                 "monoisotopic_peak", "data", "missed_peaks"]

    def __init__(self, seed_peak, score, charge, theoretical, experimental, data=None,
                 missed_peaks=0, **kwargs):
        self.seed_peak = seed_peak
        self.score = score
        self.charge = charge
        self.experimental = experimental
        self.theoretical = theoretical
        self.monoisotopic_peak = experimental[0]
        self.data = data
        self.missed_peaks = missed_peaks

    def clone(self) -> 'IsotopicFitRecord':
        """Make a shallow copy of this object.

        Returns
        -------
        IsotopicFitRecord
        """
        return self.__class__(
            self.seed_peak, self.score, self.charge, self.theoretical, self.experimental, self.data, self.missed_peaks)

    def __reduce__(self):
        return self.__class__, (
            self.seed_peak, self.score, self.charge, self.theoretical, self.experimental, self.data, self.missed_peaks)

    def __eq__(self, other):
        val = (self.score == other.score and
               self.charge == other.charge and
               self.experimental == other.experimental and
               self.theoretical == other.theoretical)
        if self.data is not None or other.data is not None:
            val = val and (self.data == other.data)
        return val

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.score < other.score

    def __gt__(self, other):
        return self.score > other.score

    def __hash__(self):
        return hash((self.monoisotopic_peak.mz, self.charge))

    def __iter__(self):
        yield self.score
        yield self.charge
        yield self.experimental
        yield self.theoretical

    @property
    def npeaks(self):
        """The number of experimental peaks covered in the fit.

        Returns
        -------
        int
        """
        return len(self.experimental)

    def __repr__(self):
        return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (
            self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)


class FitSelectorBase(Base):
    """An object that controls the filtering and
    selection of IsotopicFitRecord

    Attributes
    ----------
    minimum_score : int
        The minimum score needed to be a candidate for selection. If the
        FitSelector is `minimizing` it is the maximal score to be a candidate.
    """

    minimum_score = 0

    def __init__(self, minimum_score=0):
        self.minimum_score = minimum_score

    def best(self, results):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        return self.best(*args, **kwargs)

    def reject(self, fit):
        raise NotImplementedError()

    def reject_score(self, score):
        raise NotImplementedError()

    def is_maximizing(self):
        return False


class MinimizeFitSelector(FitSelectorBase):
    """A FitSelector which tries to minimize the score of the best fit."""

    def best(self, results):
        """Returns the IsotopicFitRecord with the smallest score

        Parameters
        ----------
        results : list of IsotopicFitRecord
            List of isotopic fits to select the most optimal case from

        Returns
        -------
        IsotopicFitRecord
            The most optimal fit
        """
        return min(results, key=operator.attrgetter("score"))

    def reject(self, fit):
        """Decide whether the fit should be discarded for having too
        large a score. Compares against :attr:`minimum_score`

        Parameters
        ----------
        fit : IsotopicFitRecord

        Returns
        -------
        bool
        """
        return fit.score > self.minimum_score

    def reject_score(self, score):
        return score > self.minimum_score

    def is_maximizing(self):
        """Returns that this is *not* a maximizing selector

        Returns
        -------
        False
        """
        return False


class MaximizeFitSelector(FitSelectorBase):
    """A FitSelector which tries to maximize the score of the best fit."""

    def best(self, results):
        """Returns the IsotopicFitRecord with the largest score

        Parameters
        ----------
        results : list of IsotopicFitRecord
            List of isotopic fits to select the most optimal case from

        Returns
        -------
        IsotopicFitRecord
            The most optimal fit
        """
        return max(results, key=operator.attrgetter("score"))

    def reject(self, fit):
        """Decide whether the fit should be discarded for having too
        small a score. Compares against :attr:`minimum_score`

        Parameters
        ----------
        fit : IsotopicFitRecord

        Returns
        -------
        bool
        """
        return fit.score < self.minimum_score

    def reject_score(self, score):
        return score < self.minimum_score

    def is_maximizing(self):
        """Returns that this *is* a maximizing selector

        Returns
        -------
        False
        """
        return True


class IsotopicFitterBase(Base):
    """A base class for Isotopic Pattern Fitters, objects
    which given a set of experimental peaks and a set of matching
    theoretical peaks, returns a fit score describing how good the
    match is.

    An IsotopicFitter may be optimal when the score is small (minimizing)
    or when the score is large (maximizing), and the appropriate
    :class:`FitSelectorBase` type will be used for :attr:`select`. This
    will also be reflected by :meth:`is_maximizing`.
    """

    def __init__(self, score_threshold=0.5):
        self.select = MinimizeFitSelector(score_threshold)

    def evaluate(self, peaklist, observed, expected, **kwargs):
        """Evaluate a pair of peak lists for goodness-of-fit.

        Parameters
        ----------
        peaklist : :class:`~.PeakSet`
            The full set of all experimental peaks
        observed : list
            The list of experimental peaks that are part of this fit
        expected : list
            The list of theoretical peaks that are part of this fit
        **kwargs

        Returns
        -------
        float
            The score
        """
        return NotImplemented

    def _evaluate(self, peaklist, observed, expected, **kwargs):
        return self.evaluate(peaklist, observed, expected, **kwargs)

    def __call__(self, *args, **kwargs):
        """Invokes :meth:`evaluate`

        Parameters
        ----------
        *args
            Forwarded to :meth:`evaluate`
        **kwargs
            Forwarded to :meth:`evaluate`

        Returns
        -------
        float
            The score
        """
        return self.evaluate(*args, **kwargs)

    def reject(self, fit):
        """Test whether this fit is too poor to be used

        Parameters
        ----------
        fit : :class:`~IsotopicFitRecord`
            The fit to test

        Returns
        -------
        bool
        """
        return self.select.reject(fit)

    def is_maximizing(self):
        """Whether or not this fitter's score gets better as it grows

        Returns
        -------
        bool
            Whether or not this fitter is a maximizing fitter
        """
        return self.select.is_maximizing()

    def configure(self, deconvoluter, **kwargs):
        return self


class GTestFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using a G-test

    .. math::
        G = 2\sum_i^n{o_i * ({log}o_i - {log}e_i)}

    where :math:`o_i` is the intensity of the ith experimental peak
    and :math:`e_i` is the intensity of the ith theoretical peak.
    """

    def evaluate(self, peaklist, observed, expected, **kwargs):
        g_score = 2 * sum([obs.intensity * np.log(
            obs.intensity / theo.intensity) for obs, theo in zip(observed, expected)])
        return g_score


g_test = GTestFitter()


class ScaledGTestFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using a G-test after normalizing the
    list of experimental and theoretical peaks to both sum to 1.

    .. math::
        G = 2\sum_i^n{o_i * ({log}o_i - {log}e_i)}

    where :math:`o_i` is the intensity of the ith experimental peak
    and :math:`e_i` is the intensity of the ith theoretical peak.
    """

    def evaluate(self, peaklist, observed, expected, **kwargs):
        total_observed = sum(p.intensity for p in observed)
        total_expected = sum(p.intensity for p in expected)
        total_expected += eps
        normalized_observed = [obs.intensity /
                               total_observed for obs in observed]
        normalized_expected = [theo.intensity /
                               total_expected for theo in expected]
        g_score = 2 * sum([obs * np.log(obs / theo) for obs, theo in zip(
            normalized_observed, normalized_expected)])
        return g_score


g_test_scaled = ScaledGTestFitter()


class ChiSquareFitter(IsotopicFitterBase):

    def evaluate(self, peaklist, observed, expected, **kwargs):
        score = sum([(obs.intensity - theo.intensity)**2 / theo.intensity
                     for obs, theo in zip(observed, expected)])
        return score


chi_sqr_test = ChiSquareFitter()


class LeastSquaresFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using least squares coefficient of
    determination :math:`R^2`.

    .. math::
        {\hat e_i} &= e_i / max(e)

        {\hat t_i} &= t_i / max(t)

        {\hat t} &= \sum_i^n {\hat t_i}^2

        R^2 &= \frac{1}{{\hat t}}\sum_i^n ({\hat e_i} - {\hat t_i})^2

    where :math:`e_i` is the ith experimental peak intensity and :math:`t_i`
    is the ith theoretical peak intensity
    """

    def evaluate(self, peaklist, observed, expected, **kwargs):
        exp_max = max(p.intensity for p in observed)
        theo_max = max(p.intensity for p in expected)

        sum_of_squared_errors = 0
        sum_of_squared_theoreticals = 0

        for e, t in zip(observed, expected):
            normed_expr = e.intensity / exp_max
            normed_theo = t.intensity / theo_max
            sum_of_squared_errors += (normed_theo - normed_expr) ** 2
            sum_of_squared_theoreticals += normed_theo ** 2
        return sum_of_squared_errors / sum_of_squared_theoreticals


least_squares = LeastSquaresFitter()


class MSDeconVFitter(IsotopicFitterBase):
    """An implementation of the scoring function used in :title-reference:`MSDeconV`

    References
    ----------
    Liu, X., Inbar, Y., Dorrestein, P. C., Wynne, C., Edwards, N., Souda, P., …
    Pevzner, P. A. (2010). Deconvolution and database search of complex tandem
    mass spectra of intact proteins: a combinatorial approach. Molecular & Cellular
    Proteomics : MCP, 9(12), 2772–2782. https://doi.org/10.1074/mcp.M110.002766
    """

    def __init__(self, minimum_score=10, mass_error_tolerance=0.02):
        self.select = MaximizeFitSelector()
        self.select.minimum_score = minimum_score
        self.mass_error_tolerance = mass_error_tolerance

    def calculate_minimum_signal_to_noise(self, observed):
        snr = 0
        n = 0
        for obs in observed:
            if obs.signal_to_noise < 1:
                continue
            snr += obs.signal_to_noise
            n += 1
        return (snr / n) * 0.05

    def reweight(self, obs, theo, obs_total, theo_total):
        norm_obs = obs.intensity / obs_total
        norm_theo = theo.intensity / theo_total
        return norm_obs * np.log(norm_obs / norm_theo)

    def score_peak(self, obs, theo, mass_error_tolerance=0.02, minimum_signal_to_noise=1):
        if obs.signal_to_noise < minimum_signal_to_noise:
            return 0.

        mass_error = np.abs(obs.mz - theo.mz)

        if mass_error <= mass_error_tolerance:
            mass_accuracy = 1 - mass_error / mass_error_tolerance
        else:
            mass_accuracy = 0

        if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
            abundance_diff = 1 - \
                ((theo.intensity - obs.intensity) / obs.intensity)
        elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
            abundance_diff = np.sqrt(
                1 - ((obs.intensity - theo.intensity) / obs.intensity))
        else:
            abundance_diff = 0.
        score = np.sqrt(theo.intensity) * mass_accuracy * abundance_diff
        return score

    def evaluate(self, peaklist, observed, expected, **kwargs):
        score = 0
        for obs, theo in zip(observed, expected):
            inc = self.score_peak(obs, theo, self.mass_error_tolerance, 1)
            score += inc
        return score


class PenalizedMSDeconVFitter(IsotopicFitterBase):
    r"""An Isotopic Fitter which uses the :class:`MSDeconVFitter` score
    weighted by 1 - :attr:`penalty_factor` * :class:`ScaledGTestFitter` score

    .. math::
        S(e, t) = M(e, t) * (1 - G(e, t))

    where :math:`e` is the experimental peak list and :math:`t` is the theoretical
    peak list
    """

    def __init__(self, minimum_score=10, penalty_factor=1., mass_error_tolerance=0.02):
        self.select = MaximizeFitSelector(minimum_score)
        self.msdeconv = MSDeconVFitter(mass_error_tolerance=mass_error_tolerance)
        self.penalizer = ScaledGTestFitter()
        self.penalty_factor = penalty_factor

    def evaluate(self, peaklist, observed, expected, **kwargs):
        score = self.msdeconv.evaluate(peaklist, observed, expected)
        penalty = abs(self.penalizer.evaluate(peaklist, observed, expected))
        return score * (1 - penalty * self.penalty_factor)


def decon2ls_chisqr_test(peaklist, observed, expected, **kwargs):
    fit_total = 0
    sum_total = 0
    for obs, theo in zip(observed, expected):
        intensity_diff = obs.intensity - theo.intensity
        fit_total += (intensity_diff ** 2) / (theo.intensity + obs.intensity)
        sum_total += theo.intensity * obs.intensity
    return fit_total / (sum_total + 0.01)


class InterferenceDetection(object):

    def __init__(self, peaklist):
        self.peaklist = peaklist

    def __call__(self, experimental_peaks, lower=None, upper=None):
        return self.detect_interference(experimental_peaks, lower, upper)

    def detect_interference(self, experimental_peaks, lower=None, upper=None):
        min_peak = experimental_peaks[0]
        max_peak = experimental_peaks[-1]

        if lower is None:
            try:
                lower = min_peak.mz - min_peak.full_width_at_half_max
            except AttributeError:
                lower = min_peak.mz
        if upper is None:
            try:
                upper = max_peak.mz + max_peak.full_width_at_half_max
            except AttributeError:
                upper = max_peak.mz

        region = self.peaklist.between(lower, upper)

        included_intensity = sum(p.intensity for p in experimental_peaks)
        region_intensity = sum(p.intensity for p in region)
        if region_intensity == 0:
            return 0.

        score = 1 - (included_intensity / region_intensity)
        return score


class DistinctPatternFitter(IsotopicFitterBase):

    def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):
        self.select = MinimizeFitSelector(minimum_score)
        self.interference_detector = None
        self.g_test_scaled = ScaledGTestFitter()
        self.peak_count_scale = peak_count_scale
        self.domain_scale = domain_scale

    def evaluate(self, peaklist, experimental, theoretical, **kwargs):
        npeaks = float(len(experimental))
        if self.interference_detector is None:
            self.interference_detector = InterferenceDetection(peaklist)

        score = self.g_test_scaled(peaklist, experimental, theoretical)
        score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
            npeaks * self.peak_count_scale)) * self.domain_scale
        return score


def percentile(N, percent):
    if not N:
        return None
    k = (len(N) - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return N[int(k)]
    d0 = N[int(f)] * (c - k)
    d1 = N[int(c)] * (k - f)
    return d0 + d1


class DotProductFitter(IsotopicFitterBase):
    def evaluate(self, peaklist, observed, expected, **kwargs):
        total = 0
        for e, t in zip(observed, expected):
            total += e.intensity * t.intensity
        return total


try:
    _has_c = True
    _IsotopicFitRecord = IsotopicFitRecord
    _LeastSquaresFitter = LeastSquaresFitter
    _MSDeconVFitter = MSDeconVFitter
    _ScaledGTestFitter = ScaledGTestFitter
    _PenalizedMSDeconVFitter = PenalizedMSDeconVFitter
    _DistinctPatternFitter = DistinctPatternFitter
    _DotProductFitter = DotProductFitter

    from ._c.scoring import (
        IsotopicFitRecord, LeastSquaresFitter, MSDeconVFitter,
        ScaledGTestFitter, PenalizedMSDeconVFitter, DistinctPatternFitter,
        DotProductFitter, FunctionScorer)
except ImportError as e:
    _has_c = False

msdeconv = MSDeconVFitter()
least_squares = LeastSquaresFitter()
g_test_scaled = ScaledGTestFitter()
penalized_msdeconv = PenalizedMSDeconVFitter()
distinct_pattern_fitter = DistinctPatternFitter()


class MassShiftSupportPostProcessorBase(object):

    def __init__(self, shifts=None):
        if shifts is None:
            shifts = []
        self.shifts = shifts
        self._fits = None

    @property
    def fits(self):
        return self._fits

    @fits.setter
    def fits(self, value):
        self._fits = value

    def reweight(self, fit, shift):
        pass



import os
import glob
import re
import sys

from collections import defaultdict, OrderedDict
from typing import Iterator

import numpy as np
from ms_deisotope.data_source.metadata.data_transformation import DataProcessingInformation, ProcessingMethod
from ms_deisotope.data_source.metadata.instrument_components import InstrumentInformation, instrument_models
from ms_deisotope.data_source.metadata.software import software_names
from ms_deisotope.data_source.scan.scan_iterator import _ScanIteratorImplBase

from ms_deisotope.utils import Base
from ms_deisotope.data_source.common import (
    Scan, ActivationInformation, PrecursorInformation,
    ChargeNotProvided, IsolationWindow)
from ms_deisotope.data_source.scan.loader import ScanDataSource, RandomAccessScanSource, ScanFileMetadataBase
from ms_deisotope.data_source.scan.mobility_frame import IonMobilityFrame, IonMobilitySourceRandomAccessFrameSource
from ms_deisotope.data_source.metadata.scan_traits import (
    ScanAcquisitionInformation, ScanWindow, ScanEventInformation,
    ion_mobility_drift_time)
from ms_deisotope.data_source.metadata.activation import HCD

from . import (MassLynxRawInfoReader,
               MassLynxRawScanReader,
               MassLynxRawChromatogramReader,
               MassLynxRawDefs,
               MassLynxParameters,
               MassLynxLockMassProcessor,
               libload
              )

# Cyclic arrival time = (bin number*pusher period*pushesperbin) + ADC start delay


waters_id_pattern = re.compile(r"function=(\d+) process=(\d+) scan=(\d+)")
waters_frame_id_pattern = re.compile(r"function=(\d+) process=(\d+) startScan=(\d+) endScan=(\d+)")


def is_waters_raw_dir(path):
    """Detect whether or not the file referenced by ``path``
    is a Waters RAW directory.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`bool`:
        Whether or not the file is a Waters RAW directory.
    """
    if not libload.proxy._is_loaded():
        try:
            libload.register_dll()
        except ImportError:
            return False
    try:
        MassLynxRawLoader(path)
        return True
    except Exception:   # pylint: disable=broad-except
        return False


def infer_reader(path):
    """If the file referenced by ``path`` is a Waters RAW
    directory, return the callable (:class:`MassLynxRawLoader`) to
    open it, otherwise raise an exception.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`type`:
        The type to use to open the file

    Raises
    ------
    :class:`ValueError`:
        If the file is not a Waters RAW file
    """
    if is_waters_raw_dir(path):
        return MassLynxRawLoader
    raise ValueError("Not Waters Raw Directory")


determine_if_available = libload.determine_if_available


class Cycle(Base):
    __slots__ = ("function", "block", "start_scan", "end_scan", "index", "id")

    def __init__(self, function=None, block=None, start_scan=None, end_scan=None, index=None, id=None):
        self.function = function
        self.block = block
        self.start_scan = start_scan
        self.end_scan = end_scan
        self.index = index
        self.id = id

    def __str__(self):
        return self.id


class IndexEntry(Base):
    __slots__ = ('function', 'process', 'block', 'scan', 'index', 'id')

    def __init__(self, function=None, process=None, block=None, scan=None, index=None, id=None):
        self.function = function
        self.process = process
        self.block = block
        self.scan = scan
        self.index = index
        self.id = id

    def __str__(self):
        return self.id


class WatersMSECycleSourceMixin(IonMobilitySourceRandomAccessFrameSource):
    def _frame_id(self, data) -> str:
        return data.id

    def _frame_index(self, data) -> int:
        return data.index

    def _frame_time(self, data) -> float:
        scan = self.get_scan_by_index(data.start_scan)
        return scan.scan_time

    def _frame_ms_level(self, data) -> int:
        scan = self.get_scan_by_index(data.start_scan)
        return scan.ms_level

    def _frame_start_scan_index(self, data) -> int:
        return data.start_scan

    def _frame_end_scan_index(self, data) -> int:
        return data.end_scan

    def _frame_precursor_information(self, data) -> PrecursorInformation:
        scan = self.get_scan_by_index(data.start_scan)
        return scan.precursor_information

    def _frame_activation(self, data) -> ActivationInformation:
        scan = self.get_scan_by_index(data.start_scan)
        return scan.activation

    def _frame_isolation_window(self, data) -> IsolationWindow:
        scan = self.get_scan_by_index(data.start_scan)
        return scan.isolation_window

    def _frame_polarity(self, data) -> int:
        scan = self.get_scan_by_index(data.start_scan)
        return scan.polarity

    def get_frame_by_id(self, id: str) -> IonMobilityFrame:
        index = self._cycle_id_to_index[id]
        return self.get_frame_by_index(index)

    def get_frame_by_index(self, index: int) -> IonMobilityFrame:
        cycle = self.cycle_index[index]
        return self._make_frame(cycle)

    def get_frame_by_time(self, time: float) -> IonMobilityFrame:
        lo = 0
        hi = len(self.cycle_index)

        best_match = None
        best_error = float('inf')

        if time == float('inf'):
            return self.get_frame_by_index(self.cycle_index[-1].index)

        while hi != lo:
            mid = (hi + lo) // 2
            sid = self.cycle_index[mid]
            frame = self.get_frame_by_index(sid.index)
            if not self._validate_frame(frame):
                sid = self.cycle_index[mid - 1]
                frame = self.get_frame_by_index(sid.index)
                if not self._validate_frame(frame):
                    sid = self.cycle_index[mid - 2]
                    frame = self.get_frame_by_index(sid.index)

            frame_time = frame.time
            err = abs(frame_time - time)
            if err < best_error:
                best_error = err
                best_match = frame
            if frame_time == time:
                # Rewind here
                i = frame.index - 1
                while i >= 0:
                    prev_frame = self.get_frame_by_index(i)
                    if prev_frame.time == frame.time:
                        frame = prev_frame
                        i -= 1
                    else:
                        break
                return frame
            elif (hi - lo) == 1:
                # Rewind here
                frame = best_match
                i = frame.index - 1
                while i >= 0:
                    prev_frame = self.get_frame_by_index(i)
                    if prev_frame.time == frame.time:
                        frame = prev_frame
                        i -= 1
                    else:
                        break
                return frame
            elif frame_time > time:
                hi = mid
            else:
                lo = mid

    def _validate_frame(self, data) -> bool:
        return True

    def _make_frame(self, data) -> IonMobilityFrame:
        return IonMobilityFrame(data, self)

    def _cache_frame(self, frame: IonMobilityFrame):
        pass

    def _default_frame_iterator(self, start_index: int=None) -> Iterator['Cycle']:
        if start_index is None:
            start_index = 0
        for i in range(start_index, len(self.cycle_index)):
            yield self.cycle_index[i]

    def make_frame_iterator(self, iterator=None, grouped=False, **kwargs) -> _ScanIteratorImplBase:
        from ms_deisotope.data_source.scan.scan_iterator import (
            _SingleScanIteratorImpl, _GroupedScanIteratorImpl, MSEIterator)
        if iterator is None:
            iterator = self._default_frame_iterator()

        if grouped == 'mse':
            strategy = MSEIterator(
                iterator, self._make_frame,
                self._validate_frame, self._cache_frame,
                low_energy_config=self.low_energy_function,
                lock_mass_config=self.lockmass_function, **kwargs)
        elif grouped:
            strategy = _GroupedScanIteratorImpl(
                iterator, self._make_frame, self._validate_frame, self._cache_frame, **kwargs)
        else:
            strategy = _SingleScanIteratorImpl(
                iterator, self._make_frame, self._validate_frame, self._cache_frame, **kwargs)
        return strategy

    def start_from_frame(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped='mse', **kwargs):
        if scan_id is not None:
            frame = self.get_frame_by_id(scan_id)
            start_index = frame.index
        elif index is not None:
            start_index = index
        elif rt is not None:
            frame = self.get_frame_by_time(rt)
            start_index = frame.index
        if require_ms1 and grouped != 'mse':
            while start_index != 0:
                frame = self.get_frame_by_index(start_index)
                if frame.ms_level > 1:
                    start_index -= 1
                else:
                    break
        self._producer = self.make_frame_iterator(
            self._default_frame_iterator(start_index), grouped=grouped, **kwargs)
        return self


class WatersMassLynxScanSource(ScanDataSource):
    default_isolation_width = 0.0

    def _make_scan(self, data):
        return Scan(data, self)

    def _scan_time(self, scan):
        if scan.block >= 0:
            return self.info_reader.GetRetentionTime(scan.function, scan.block)
        else:
            return self.info_reader.GetRetentionTime(scan.function, scan.scan)

    def _drift_time(self, scan):
        if scan.block >= 0:
            return self.info_reader.GetDriftTime(scan.function, scan.scan)
        else:
            return None

    def _scan_id(self, scan):
        return scan.id

    def _scan_index(self, scan):
        return scan.index

    def _ms_level(self, scan):
        ms_level, _scan_type = self._translate_function_type(scan.function)
        return ms_level

    def _scan_arrays(self, scan):
        if scan.block >= 0:
            mz, inten = self.scan_reader.ReadDriftScan(
                scan.function, scan.block, scan.scan)
        else:
            mz, inten = self.scan_reader.ReadScan(scan.function, scan.scan)
        return np.array(mz).astype(float), np.array(inten).astype(float)

    def _precursor_information(self, scan):
        if self._ms_level(scan) == 1:
            return None
        set_mass_str = self.info_reader.GetScanItem(
            scan.function,
            scan.block if scan.block >= 0 else scan.scan,
            MassLynxRawDefs.MassLynxScanItem.SET_MASS.value
        )
        if set_mass_str:
            set_mass = float(set_mass_str)
        else:
            set_mass = 0.0
        if set_mass == 0:
            lower_bound, upper_bound = self.info_reader.GetAcquisitionMassRange(
                scan.function)
            set_mass = (lower_bound + upper_bound) / 2.
        pinfo = PrecursorInformation(
            set_mass, 0, ChargeNotProvided, source=self, product_scan_id=scan.id)
        return pinfo

    def _isolation_window(self, scan):
        if self._ms_level(scan) == 1:
            return None
        set_mass_str = self.info_reader.GetScanItem(
            scan.function,
            scan.block if scan.block >= 0 else scan.scan,
            MassLynxRawDefs.MassLynxScanItem.SET_MASS.value
        )
        if set_mass_str:
            set_mass = float(set_mass_str)
        else:
            set_mass = 0.0
        if set_mass == 0:
            lower_bound, upper_bound = self.info_reader.GetAcquisitionMassRange(
                scan.function)
            set_mass = (lower_bound + upper_bound) / 2.
            lower_bound_offset = upper_bound_offset = upper_bound - set_mass
        else:
            if self.default_isolation_width:
                lower_bound_offset = upper_bound_offset = self.default_isolation_width
            else:
                lower_bound, upper_bound = self.info_reader.GetAcquisitionMassRange(
                    scan.function)
                lower_bound_offset = set_mass - lower_bound
                upper_bound_offset = upper_bound - set_mass
        return IsolationWindow(
            lower_bound_offset, set_mass, upper_bound_offset)

    def _is_profile(self, scan):
        return self.info_reader.IsContinuum(scan.function)

    def _scan_title(self, scan):
        return self._scan_id(scan)

    def _acquisition_information(self, scan):
        scan_window = ScanWindow(
            *self.info_reader.GetAcquisitionMassRange(scan.function))
        scan_time = self._scan_time(scan)
        drift_time = self._drift_time(scan)
        event = ScanEventInformation(scan_time, [scan_window], traits={
            'preset scan configuration': scan.function + 1,
        })
        if drift_time is not None:
            event._ion_mobility.add_ion_mobility(
                ion_mobility_drift_time, drift_time)
        return ScanAcquisitionInformation('no combination', [event])

    def _polarity(self, scan):
        mode = self.info_reader.GetIonMode(scan.function)
        s = self.info_reader.GetIonModeString(mode)
        if s.endswith('+'):
            return 1
        elif s.endswith('-'):
            return -1
        raise ValueError("Unknown Ion Mode %r" % (s, ))
        # return 1

    def _activation(self, scan):
        energy_str = self.info_reader.GetScanItem(
            scan.function, scan.block if scan.block >= 0 else scan.scan, MassLynxRawDefs.MassLynxScanItem.COLLISION_ENERGY)
        if energy_str:
            energy = float(energy_str)
            return ActivationInformation(HCD, energy)


class MassLynxMetadata(ScanFileMetadataBase):
    def instrument_configuration(self):
        inst_model = instrument_models['Waters instrument model']
        return [InstrumentInformation(1, [], inst_model)]

    def file_description(self):
        finfo = super().file_description()
        return finfo

    def data_processing(self):
        return super().data_processing()

    def software_list(self):
        sw = software_names['MassLynx']
        return [sw]


class MassLynxRawLoader(RandomAccessScanSource, WatersMassLynxScanSource, WatersMSECycleSourceMixin, MassLynxMetadata):
    def __init__(self, raw_path, lockmass_config=None, default_isolation_width=0.0, **kwargs):
        if sys.version_info.major == 2:
            if not isinstance(raw_path, str):
                raw_path = str(raw_path)
        self.source_file = raw_path

        self.info_reader = MassLynxRawInfoReader.MassLynxRawInfoReader(
            self.source_file)
        self.chrom_reader = MassLynxRawChromatogramReader.MassLynxRawChromatogramReader(
            self.source_file)
        self.scan_reader = MassLynxRawScanReader.MassLynxRawScanReader(
            self.source_file)
        self.lockmass_processor = MassLynxLockMassProcessor.MassLynxLockMassProcessor()
        self.lockmass_processor.SetRawData(self.scan_reader)
        self.configure_lockmass(lockmass_config)
        self.index = []
        self.cycle_index = []
        self._cycle_id_to_index = {}
        self._producer = None
        self.initialize_scan_cache()

        self._build_function_index()
        self._build_scan_index()
        self.lockmass_function = self.function_index_list[-1] + 1
        self.low_energy_function = self.function_index_list[0] + 1
        self.default_isolation_width = default_isolation_width or 0.0

        self._producer = self.make_frame_iterator()

    def configure_lockmass(self, lockmass_config=None):
        if lockmass_config is None:
            return
        if isinstance(lockmass_config, (list, tuple)):
            if len(lockmass_config) == 1:
                mass = float(lockmass_config[0])
                tolerance = 0.25
            else:
                mass = float(lockmass_config[0])
                tolerance = float(lockmass_config[1])
        else:
            mass = float(lockmass_config)
            tolerance = 0.25

        params = MassLynxParameters.MassLynxParameters()
        params.Set(MassLynxRawDefs.LockMassParameter.MASS, mass)
        params.Set(MassLynxRawDefs.LockMassParameter.TOLERANCE, tolerance)
        self.lockmass_processor.SetParameters(params)
        if self.lockmass_processor.CanLockMassCorrect():
            self.lockmass_processor.LockMassCorrect()

    def __repr__(self):
        return "MassLynxRawLoader(%r)" % (self.source_file)

    # Vendor data access support methods
    def _read_header_properties(self):
        path = os.path.join(self.source_file, "_HEADER.TXT")
        self.header_properties = {}
        with open(path, 'rt') as fh:
            for line in fh:
                if not line.startswith("$$ "):
                    continue
                name, value = line[3:].split(": ")
                self.header_properties[name] = value.strip()

    def _build_function_index(self):
        num_spectra = 0
        function_file_path_by_number = {}
        function_index_list = []
        for dat in glob.glob(os.path.join(self.source_file, "_FUNC*.DAT")):
            fnum = int(os.path.basename(dat).replace(
                "_func", '').replace(".dat", ""))
            function_file_path_by_number[fnum - 1] = dat
            function_index_list.append(fnum - 1)
            num_spectra += self.info_reader.GetScansInFunction(fnum - 1)
        function_index_list.sort()
        self.num_spectra = num_spectra
        self.function_index_list = function_index_list
        self.function_file_path_by_number = function_file_path_by_number
        self.ion_mobility_by_function_index = OrderedDict()
        self.times_by_function_index = OrderedDict()
        self.tic_by_function_index = OrderedDict()
        self.sonar_enabled_by_function_index = OrderedDict()
        self.tic_by_function_index = OrderedDict()
        self.times_by_function_index = OrderedDict()
        for fnum, dat in function_file_path_by_number.items():
            cdt = dat.replace(".dat", ".cdt")
            self.ion_mobility_by_function_index[fnum] = False
            if os.path.exists(cdt) and self.info_reader.GetDriftScanCount(fnum) > 0:
                self.ion_mobility_by_function_index[fnum] = True
                self.sonar_enabled_by_function_index[fnum] = int(
                    self.info_reader.GetScanItem(
                        fnum, 0, MassLynxRawDefs.MassLynxScanItem.SONAR_ENABLED.value))
            time, tic = self.chrom_reader.ReadTIC(fnum)
            self.tic_by_function_index[fnum] = np.array(tic)
            self.times_by_function_index[fnum] = np.array(time)
        self._read_header_properties()

    def _translate_function_type(self, fnum):
        fcode = self.info_reader.GetFunctionType(fnum)
        name = self.info_reader.GetFunctionTypeString(fcode)
        if name in ("MSMS", "TOFD", "MS2", ):
            return 2, "MSn spectrum"
        elif name in ("MS", "TOF MS", ):
            return 1, "MS1 spectrum"
        else:
            raise ValueError("Unknown function type")

    def _build_scan_index(self):
        function_and_scan_by_rt = []
        # self.scan_time_to_function_block_map = defaultdict(list)
        num_scans_in_block = 0
        for fnum in self.function_index_list:
            _ms_level, spectrum_type = self._translate_function_type(fnum)
            if spectrum_type not in ("MSn spectrum", "MS1 spectrum"):
                continue
            scan_count = self.info_reader.GetScansInFunction(fnum)
            if self.ion_mobility_by_function_index[fnum]:
                num_scans_in_block = self.info_reader.GetDriftScanCount(fnum)
                for i in range(scan_count):
                    key = self.info_reader.GetRetentionTime(fnum, i)
                    # self.scan_time_to_function_block_map[key *
                    #                                      60].append((fnum, i))
                    function_and_scan_by_rt.append((key, (fnum, i)))
            else:
                for i in range(scan_count):
                    function_and_scan_by_rt.append(
                        (self.info_reader.GetRetentionTime(fnum, i), (fnum, i)))
        function_and_scan_by_rt.sort(key=lambda x: x[0])
        self.index = []
        self.cycle_index = []
        self._cycle_id_to_index = {}
        self.function_blocks = defaultdict(list)
        for _rt, (fnum, i) in function_and_scan_by_rt:
            if self.ion_mobility_by_function_index[fnum]:
                # num_scans_in_block = self.info_reader.GetDriftScanCount(fnum)
                block_start = len(self.index)
                for j in range(num_scans_in_block):
                    self.index.append(IndexEntry())
                    ie = self.index[-1]
                    ie.function = fnum
                    ie.process = 0
                    ie.block = i
                    ie.scan = j
                    ie.index = len(self.index) - 1
                    ie.id = "function=%d process=%d scan=%d" % (ie.function + 1, ie.process,
                                                                num_scans_in_block * ie.block + ie.scan + 1)
                block_end = len(self.index)
                self.function_blocks[fnum].append((block_start, block_end))
                cyc = Cycle(
                    fnum, i, block_start, block_end,
                    len(self.cycle_index),
                    id="function=%d process=0 startScan=%d endScan=%d" % (
                        fnum + 1,
                        num_scans_in_block * i + 1,
                        num_scans_in_block * i + num_scans_in_block,
                    ))
                self.cycle_index.append(cyc)
                self._cycle_id_to_index[cyc.id] = len(self.cycle_index) - 1
            else:
                block_start = len(self.index)
                self.index.append(IndexEntry())
                ie = self.index[-1]
                ie.function = fnum
                ie.process = 0
                ie.block = -1
                ie.scan = i
                ie.index = len(self.index) - 1
                ie.id = "function=%d process=%d scan=%d" % (ie.function + 1, ie.process,
                                                            ie.scan + 1)
                block_end = len(self.index)
                self.function_blocks[fnum].append((block_start, block_end))
                cyc = Cycle(
                    ie.function, ie.block, block_start, block_end,
                    len(self.cycle_index),
                    id="function=%d process=0 startScan=%d endScan=%d" % (
                        fnum + 1,
                        i + 1,
                        i + 2,
                    ))
                self.cycle_index.append(cyc)
                self._cycle_id_to_index[cyc.id] = len(self.cycle_index) - 1

    # RandomAccessScanSource methods
    def __len__(self):
        return len(self.index)

    def get_scan_by_index(self, index: int) -> Scan:
        ie = self.index[index]
        if ie.id in self._scan_cache:
            return self._scan_cache[ie.id]
        scan = self._make_scan(ie)
        self._scan_cache[ie.id] = scan
        return scan

    def get_scan_by_id(self, scan_id: str) -> Scan:
        match = waters_id_pattern.search(scan_id)
        if not match:
            raise KeyError(scan_id)
        fnum1, _proc, scan = tuple(map(int, match.groups()))
        fnum = fnum1 - 1
        intervals = self.function_blocks[fnum]
        for interval in intervals:
            ie = self.index[interval[1] - 1]
            snum = int(waters_id_pattern.search(ie.id).group(3))
            if scan <= snum:
                ie = self.index[interval[0]]
                snum = int(waters_id_pattern.search(ie.id).group(3))
                if snum <= scan:
                    for i in range(interval[0], interval[1]):
                        ie = self.index[i]
                        if ie.id == scan_id:
                            return self.get_scan_by_index(ie.index)
                    raise KeyError(scan_id)

    def get_scan_by_time(self, time: float) -> Scan:
        lo = 0
        hi = len(self.index)

        best_match = None
        best_error = float('inf')

        if time == float('inf'):
            return self.get_scan_by_index(self.index[-1].index)

        while hi != lo:
            mid = (hi + lo) // 2
            sid = self.index[mid]
            scan = self.get_scan_by_index(sid.index)
            if not self._validate(scan):
                sid = self.index[mid - 1]
                scan = self.get_scan_by_index(sid.index)
                if not self._validate(scan):
                    sid = self.index[mid - 2]
                    scan = self.get_scan_by_index(sid.index)

            scan_time = scan.scan_time
            err = abs(scan_time - time)
            if err < best_error:
                best_error = err
                best_match = scan
            if scan_time == time:
                # Rewind here
                i = scan.index - 1
                while i >= 0:
                    prev_scan = self.get_scan_by_index(i)
                    if prev_scan.scan_time == scan.scan_time:
                        scan = prev_scan
                        i -= 1
                    else:
                        break
                return scan
            elif (hi - lo) == 1:
                # Rewind here
                scan = best_match
                i = scan.index - 1
                while i >= 0:
                    prev_scan = self.get_scan_by_index(i)
                    if prev_scan.scan_time == scan.scan_time:
                        scan = prev_scan
                        i -= 1
                    else:
                        break
                return scan
            elif scan_time > time:
                hi = mid
            else:
                lo = mid

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True, **kwargs):
        if scan_id is not None:
            scan = self.get_scan_by_id(scan_id)
            start_index = scan.index
        elif index is not None:
            start_index = index
        elif rt is not None:
            scan = self.get_scan_by_time(rt)
            start_index = scan.index
        if require_ms1:
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            current_block = self.index[start_index].block
            while start_index > 0:
                ie = self.index[start_index - 1]
                if ie.block != current_block:
                    break
                else:
                    start_index -= 1
        self._producer = self._make_pointer_iterator(start_index)
        return self

    # ScanIterator methods
    def next(self):
        return next(self._producer)

    def _make_default_iterator(self):
        return self._make_pointer_iterator()

    def _make_scan_index_producer(self, start_index=None, start_time=None):
        if start_index is not None:
            return range(start_index, len(self.index))
        elif start_time is not None:
            raise NotImplementedError()
            # start_index = self._scan_time_to_scan_number(start_time)
            # while start_index != 0:
            #     scan = self.get_scan_by_index(start_index)
            #     if scan.ms_level > 1:
            #         start_index -= 1
            #     else:
            #         break
            # return range(start_index, len(self.index))
        else:
            return range(0, len(self.index))

    def _make_pointer_iterator(self, start_index=None, start_time=None):
        iterator = self._make_scan_index_producer(start_index, start_time)
        for i in iterator:
            yield self._make_scan(self.index[i])

    def reset(self):
        self.initialize_scan_cache()
        self._producer = self.make_frame_iterator()



"""
High Level Deconvolution API
----------------------------

A high-level wrapper around the deconvolution machinery, orchestrating the
process of constructing a deconvoluter instance, performing deconvolution, and
extracting targets of interest from the result.
"""

from ms_peak_picker import FittedPeak

from ms_deisotope.peak_set import merge
from ms_deisotope.utils import DeconvolutionProcessResult
from ms_deisotope.constants import (
    ERROR_TOLERANCE,
    TRUNCATE_AFTER,
    MAX_ITERATION,
    SCALE_METHOD)

from ms_deisotope.averagine import PROTON

from .averagine_based import AveraginePeakDependenceGraphDeconvoluter
from .utils import logger, prepare_peaklist


def deconvolute_peaks(peaklist, decon_config=None,
                      charge_range=(1, 8), error_tolerance=ERROR_TOLERANCE,
                      priority_list=None, left_search_limit=1, right_search_limit=0,
                      left_search_limit_for_priorities=None, right_search_limit_for_priorities=None,
                      verbose_priorities=False, verbose=False, charge_carrier=PROTON,
                      truncate_after=TRUNCATE_AFTER, iterations=MAX_ITERATION,
                      deconvoluter_type=AveraginePeakDependenceGraphDeconvoluter,
                      retention_strategy=None,
                      use_quick_charge=False,
                      **kwargs):
    """Deconvolute a centroided mass spectrum.

    This function constructs a deconvoluter object using the ``deconvoluter_type`` argument
    and deconvolutes the input ``peaklist`` by calling its :meth:`deconvolute` method.

    If ``priority_list`` is not :const:`None`, it is expected to be an iterable of either
    tuples of (:class:`~.FittedPeak`, ``(min charge, max charge)``) pairs, or instances of
    :class:`~.PriorityTarget`. These will be passed to :meth:`targeted_deconvolution` of
    the deconvoluter.

    Parameters
    ----------
    peaklist : :class:`~.PeakSet` or list of Peak-like objects, see :func:`~.prepare_peaklist`
        The centroided mass spectrum to deconvolute.
    decon_config : dict, optional
        Parameters to use to initialize the deconvoluter instance produced by
        ``deconvoluter_type``
    charge_range : tuple of integers, optional
        The range of charge states to consider. The range is inclusive.
    error_tolerance : float, optional
        PPM error tolerance to use to match experimental to theoretical peaks
    priority_list : list, optional
        The set of peaks to target for deconvolution to be able to enforce external
        constraints on, such as selected precursors for fragmentation.
    left_search_limit : int, optional
        The maximum number of neutron shifts to search to the left  (decrease) from
        each query peak
    right_search_limit : int, optional
        The maximum number of neutron shifts to search to the right (increase) from
        each query peak
    left_search_limit_for_priorities : int, optional
        The maximum number of neutron shifts to search to the left (decrease) from
        each query peak for priority targets
    right_search_limit_for_priorities : int, optional
        The maximum number of neutron shifts to search to the right (increase) from
        each query peak for priority targets
    verbose_priorities : bool, optional
        Whether to turn on verbose mode for priority targets
    verbose : bool, optional
        Passed to the deconvoluter to enable verbose mode globally
    charge_carrier : float, optional
        The mass of the charge carrier. Defaults to |PROTON|
    truncate_after : float, optional
        The percentage of the isotopic pattern to include. Defaults to |TRUNCATE_AFTER|
    deconvoluter_type : type or callable, optional
        A callable returning a deconvoluter. Defaults to :class:`~.AveraginePeakDependenceGraphDeconvoluter`
    retention_strategy: :class:`~.PeakRetentionStrategyBase` or callable, optional
        A callable that may compute additional peaks to include in the output deconvoluted peaks.
    use_quick_charge: :class:`bool`
        Whether or not to use the :title-reference:`QuickCharge` algorithm to quickly filter
        theoretical charge states to consider for each peak.
    **kwargs
        Additional keywords included in ``decon_config``


    See Also
    --------
    :func:`ms_deisotope.peak_set.decharge`


    Notes
    -----
    If speed is an issue, consider setting `use_quick_charge` :const:`True`. This will pre-screen charge
    states that are are missing peaks supporting the :term:`monoisotopic peak` or the :term:`A+1 peak`.
    Alternatively, you may set the charge range upper bound to something reasonable for your data, such
    as the precursor ion's charge when considering a product ion spectrum.

    When you do not expect a complete isotopic pattern for large ions, as is often the case for low
    abundance FT-MS/MS it may be useful to shrink the `truncate_after` parameter from the default (|TRUNCATE_AFTER|)
    to a slightly smaller value.

    Returns
    -------
    :class:`~.DeconvolutionProcessResult`
    """
    if priority_list is None:
        priority_list = []
    if left_search_limit_for_priorities is None:
        left_search_limit_for_priorities = left_search_limit
    if right_search_limit_for_priorities is None:
        right_search_limit_for_priorities = right_search_limit

    decon_config = decon_config or {}
    decon_config.update(kwargs)
    decon_config.setdefault("use_subtraction", True)
    decon_config.setdefault("scale_method", SCALE_METHOD)
    decon_config.setdefault("use_quick_charge", use_quick_charge)

    peaklist = prepare_peaklist(peaklist)

    decon = deconvoluter_type(peaklist=peaklist, **decon_config)


    if verbose_priorities or verbose:
        decon.verbose = True

    priority_list_results = []
    for p in priority_list:
        try:
            target_info = p
            p = target_info.peak
            hinted_charge_range = target_info.charge_range_hint(charge_range)
        except AttributeError:
            hinted_charge_range = charge_range
        if not isinstance(p, FittedPeak):
            p = decon.peaklist.has_peak(p, error_tolerance)
        priority_result = decon.targeted_deconvolution(
            p, error_tolerance=error_tolerance,
            charge_range=hinted_charge_range,
            left_search_limit=left_search_limit_for_priorities,
            right_search_limit=right_search_limit_for_priorities,
            charge_carrier=charge_carrier,
            truncate_after=truncate_after)
        priority_list_results.append(priority_result)

    if verbose_priorities and not verbose:
        decon.verbose = False

    deconvoluted_peaks = decon.deconvolute(
        error_tolerance=error_tolerance, charge_range=charge_range,
        left_search_limit=left_search_limit, right_search_limit=right_search_limit,
        charge_carrier=charge_carrier, truncate_after=truncate_after,
        iterations=iterations)

    acc = []
    errors = []
    for pr in priority_list_results:
        try:
            result = pr.get()
        except Exception as e:
            result = None
            errors.append(e)
            logger.error("Could not extract a solution for %r",
                         pr.query_peak, exc_info=True)
        acc.append(result)

    priority_list_results = acc

    if retention_strategy is not None:
        retained = retention_strategy(
            decon.peaklist, peaklist, charge_range, deconvoluted_peaks)
        deconvoluted_peaks = merge(deconvoluted_peaks, retained)

    return DeconvolutionProcessResult(
        decon, deconvoluted_peaks, priority_list_results, errors)



# cython: embedsignature=True

cimport cython

from ms_deisotope.utils import uid as _uid

from collections import deque
from random import sample as _sample

from libc.stdlib cimport malloc, free

from cpython cimport PyErr_SetString
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_GetSlice, PyList_GetItem
from ms_peak_picker._c.peak_set cimport FittedPeak, PeakBase
from brainpy._c.isotopic_distribution cimport TheoreticalPeak

from ms_deisotope._c.peak_set cimport DeconvolutedPeak

cimport numpy as np
import numpy as np


np.import_array()


cdef object _zeros = np.zeros

cdef double INF = float("inf")


cdef class LCMSFeatureTreeList(object):
    def __init__(self, roots=None):
        if roots is None:
            roots = []
        self.roots = list(roots)
        self._node_id_hash = None

    @staticmethod
    cdef LCMSFeatureTreeList _create(list roots):
        cdef:
            LCMSFeatureTreeList inst
        inst = LCMSFeatureTreeList.__new__(LCMSFeatureTreeList)
        inst.roots = list(roots)
        inst._node_id_hash = None
        return inst

    cpdef LCMSFeatureTreeNodeBase _make_node(self, double time, list peaks):
        return LCMSFeatureTreeNode(time, peaks)

    cdef void _invalidate(self):
        self._node_id_hash = None

    cpdef tuple find_time(self, double time):
        cdef:
            size_t indexout
            LCMSFeatureTreeNodeBase node
        node = self._find_time(time, &indexout)
        return node, indexout

    cdef LCMSFeatureTreeNodeBase _find_time(self, double time, size_t* indexout):
        cdef:
            int lo, hi
            double rt
            LCMSFeatureTreeNodeBase node
        hi = PyList_GET_SIZE(self.roots)
        if hi == 0:
            raise ValueError()
        lo = 0
        while lo != hi:
            i = (lo + hi) // 2
            node = <LCMSFeatureTreeNodeBase>PyList_GET_ITEM(self.roots, i)
            rt = node.time
            if rt == time:
                indexout[0] = i
                return node
            elif (hi - lo) == 1:
                indexout[0] = i
                return None
            elif rt < time:
                lo = i
            elif rt > time:
                hi = i

    def _build_node_id_hash(self):
        self._node_id_hash = set()
        for node in self.unspool():
            self._node_id_hash.add(node.node_id)

    @property
    def node_id_hash(self):
        if self._node_id_hash is None:
            self._build_node_id_hash()
        return self._node_id_hash

    def insert_node(self, LCMSFeatureTreeNodeBase node):
        self._invalidate()
        cdef:
            LCMSFeatureTreeNodeBase root
            size_t i
        i = 0
        try:
            root = self._find_time(node.time, &i)
            if root is None:
                if i != 0:
                    self.roots.insert(i + 1, node)
                else:
                    slot = self.getitem(i)
                    if slot.time < node.time:
                        i += 1
                    self.roots.insert(i, node)
            else:
                root.add(node)
            return i
        except ValueError:
            self.roots.append(node)
            return 0

    def insert(self, time, peaks):
        node = self._make_node(time, peaks)
        return self.insert_node(node)

    def extend(self, iterable):
        for peaks, time in iterable:
            self.insert(time, peaks)

    def __getitem__(self, i):
        return self.roots[i]

    cdef LCMSFeatureTreeNodeBase getitem(self, size_t i):
        return <LCMSFeatureTreeNodeBase>PyList_GetItem(self.roots, i)

    def __len__(self):
        return PyList_GET_SIZE(self.roots)

    cdef size_t get_size(self):
        return PyList_GET_SIZE(self.roots)

    def __iter__(self):
        return iter(self.roots)

    def __reduce__(self):
        return self.__class__, (self.roots,)

    def clone(self, deep=False):
        return LCMSFeatureTreeList(node.clone(deep=deep) for node in self)

    def unspool(self):
        cdef:
            list out_queue, stack
            LCMSFeatureTreeNodeBase node, root
            size_t i, n
        out_queue = []
        n = self.get_size()
        for i in range(n):
            root = self.getitem(i)
            stack = [root]
            while len(stack) != 0:
                node = stack.pop()
                out_queue.append(node)
                stack.extend(node.children)
        return out_queue

    def common_nodes(self, other):
        return len(self.node_id_hash & other.node_id_hash)

    def __repr__(self):
        return "LCMSFeatureTreeList(%d nodes, %0.2f-%0.2f)" % (
            len(self), self[0].time, self[-1].time)


cdef object uid = _uid
cdef object sample = _sample


cdef class LCMSFeatureTreeNodeBase(object):

    def clone(self, deep=False):
        cdef LCMSFeatureTreeNodeBase node
        cdef list peaks
        cdef size_t i, n
        cdef PeakBase peak
        if deep:
            peaks = []
            n = self.get_members_size()
            for i in range(n):
                peak = self.getitem(i)
                if peak is not None:
                    peak = peak.clone()
                peaks.append(peak)
        else:
            peaks = list(self.members)
        node = self.__class__(self.time, peaks)
        node.node_id = self.node_id
        return node

    def _unspool_strip_children(self):
        node = self.__class__(
            self.time, list(self.members))
        yield node

    def __getstate__(self):
        return (self.time, self.members, self.node_id)

    def __setstate__(self, state):
        self.time, self.members, self.node_id = state
        self._recalculate()

    def __reduce__(self):
        return self.__class__, (self.time, []), self.__getstate__()

    def add(self, node, recalculate=True):
        if self.node_id == node.node_id:
            raise ValueError("Duplicate Node %s" % node)
        self.members.extend(node.members)
        if recalculate:
            self._recalculate()

    cpdef double _total_intensity_members(self):
        cdef:
            double total
            size_t n, i
            PeakBase peak
        total = 0.
        n = self.get_members_size()
        for i in range(n):
            peak = self.getitem(i)
            total += peak.intensity
        return total

    cpdef double max_intensity(self):
        return self._most_abundant_member.intensity

    cpdef double total_intensity(self):
        return self._total_intensity_members()

    cpdef _recalculate(self):
        self._calculate_most_abundant_member()

    cdef inline size_t get_members_size(self):
        return PyList_GET_SIZE(self.members)

    cdef PeakBase _find_most_abundant_member(self):
        cdef:
            size_t i, n
            PeakBase peak, best
            double best_intensity
        n = self.get_members_size()
        best = self.getitem(0)
        best_intensity = best.intensity
        for i in range(n):
            peak = self.getitem(i)
            if peak.intensity > best_intensity:
                best = peak
                best_intensity = best.intensity
        return best

    cpdef _calculate_most_abundant_member(self):
        cdef:
            size_t n
        n = self.get_members_size()
        if n == 1:
            self._most_abundant_member = self.getitem(0)
        else:
            if n == 0:
                self._most_abundant_member = None
            else:
                self._most_abundant_member = self._find_most_abundant_member()
        return

    cdef PeakBase getitem(self, size_t i):
        return <PeakBase>PyList_GET_ITEM(self.members, i)

    @property
    def peaks(self):
        peaks = list(self.members)
        return peaks

    def __hash__(self):
        return hash(self.uid)


cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):
    def __init__(self, time, members=None):
        if members is None:
            members = []
        self.time = time
        self.members = members
        self._most_abundant_member = None
        self._mz = 0
        self._recalculate()
        self.node_id = uid()

    cpdef _recalculate(self):
        self._calculate_most_abundant_member()
        if self._most_abundant_member is not None:
            self._mz = self._most_abundant_member.mz

    @property
    def mz(self):
        if self._mz == 0:
            if self._most_abundant_member is not None:
                self._mz = self._most_abundant_member.mz
        return self._mz

    cdef double get_mz(self):
        if self._mz == 0:
            if self._most_abundant_member is not None:
                self._mz = self._most_abundant_member.mz
        return self._mz

    cpdef bint _eq(self, LCMSFeatureTreeNode other):
        return self.members == other.members and abs(
            self.time - other.time) < 1e-4

    cpdef bint _ne(self, LCMSFeatureTreeNode other):
        return not (self._eq(other))

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return self._ne(other)
        else:
            return NotImplemented

    def __repr__(self):
        return "%s(%f, %0.4f %s|%d)" % (
            self.__class__.__name__,
            self.time, self.peaks[0].mz, "",
            len(self.members))

    cdef inline FittedPeak getpeak(self, size_t i):
        return <FittedPeak>PyList_GET_ITEM(self.members, i)


@cython.freelist(10000000)
cdef class FeatureBase(object):
    def __init__(self, nodes):
        self.nodes = LCMSFeatureTreeList(nodes)

    cdef double get_mz(self):
        return 0.0

    cdef double get_neutral_mass(self):
        return self.get_mz()

    cdef double get_start_time(self):
        return 0.0

    cdef double get_end_time(self):
        return INF

    cpdef tuple find_time(self, double time):
        cdef:
            size_t indexout
            LCMSFeatureTreeNode node
        node = self._find_time(time, &indexout)
        return node, indexout

    cdef inline LCMSFeatureTreeNode _find_time(self, double time, size_t* indexout):
        return self.nodes._find_time(time, indexout)

    cdef inline size_t get_size(self):
        return self.nodes.get_size()

    def __len__(self):
        return self.get_size()

    def __iter__(self):
        return iter(self.nodes)


cdef class LCMSFeature(FeatureBase):
    def __init__(self, nodes=None, adducts=None, used_as_adduct=None, feature_id=None):
        if nodes is None:
            nodes = []
        if adducts is None:
            adducts = []
        if used_as_adduct is None:
            used_as_adduct = []
        if feature_id is None:
            feature_id = uid()

        FeatureBase.__init__(self, nodes)
        self._total_intensity = -1
        self._mz = -1
        self._last_mz = 0.0
        self._times = None
        self._peaks = None
        self._start_time = -1
        self._end_time = -1
        self.adducts = adducts
        self.used_as_adduct = used_as_adduct
        self.feature_id = feature_id
        self._initialize_averager()

    cpdef _initialize_averager(self):
        self._peak_averager = RunningWeightedAverage._create(None)

    cpdef _update_from_averager(self):
        best_mz = self._peak_averager.current_mean
        if best_mz < 0:
            raise ValueError("best_mz < 0")
        self._last_mz = self._mz = best_mz

    cpdef _reaverage_and_update(self):
        self._peak_averager.reset()
        self._peak_averager.feed_from_feature(self)
        self._update_from_averager()

    def invalidate(self, reaverage=False):
        self._invalidate(reaverage)

    cpdef _invalidate(self, bint reaverage=False):
        self._total_intensity = -1
        self._last_mz = self._mz if self._mz != -1 else 0.
        self._mz = -1
        self._times = None
        self._peaks = None
        self._start_time = -1
        self._end_time = -1
        if reaverage:
            self._reaverage_and_update()

    @property
    def total_signal(self):
        cdef:
            double total
            size_t i, n
            LCMSFeatureTreeNode node
        if self._total_intensity == -1:
            total = 0.
            n = self.get_size()
            for i in range(n):
                node = self.getitem(i)
                total += node.total_intensity()
            self._total_intensity = total
        return self._total_intensity

    @property
    def intensity(self):
        return self.total_signal

    @property
    def neutral_mass(self):
        return self.mz

    cdef double get_mz(self):
        if self._mz == -1:
            self._reaverage_and_update()
        return self._mz

    @property
    def mz(self):
        return self.get_mz()

    @property
    def times(self):
        cdef:
            size_t i, n
            LCMSFeatureTreeNode node
            np.npy_intp knd
            np.ndarray[double, ndim=1, mode='c'] acc
        if self._times is None:
            n = self.get_size()
            knd = n
            acc = np.PyArray_ZEROS(1, &knd, np.NPY_FLOAT64, 0)
            # acc = _zeros(n)
            for i in range(n):
                node = self.getitem(i)
                acc[i] = node.time
            self._times = acc
        return self._times

    @property
    def peaks(self):
        if self._peaks is None:
            self._peaks = tuple(node.peaks for node in self.nodes)
        return self._peaks

    @property
    def start_time(self):
        if self._start_time == -1:
            self._start_time = self.nodes[0].time
        return self._start_time

    @property
    def end_time(self):
        if self._end_time == -1:
            self._end_time = self.nodes[-1].time
        return self._end_time

    cdef double get_start_time(self):
        if self._start_time == -1:
            self._start_time = self.getitem(0).time
        return self._start_time

    cdef double get_end_time(self):
        cdef:
            size_t n
        if self._end_time == -1:
            n = self.get_size()
            self._end_time = self.getitem(n - 1).time
        return self._end_time

    cpdef bint overlaps_in_time(self, FeatureBase interval):
        """Test whether or not this feature overlaps with `interval`.

        Parameters
        ----------
        interval : FeatureBase
            The feature to test spanning

        Returns
        -------
        bool:
            Whether this feature spans `interval` or the `interval` spans this feature,
            partially or one is fully contained in the other.
        """
        cdef:
            double self_start_time, self_end_time
            double other_start_time, other_end_time
            bint cond

        self_start_time = self.get_start_time()
        self_end_time = self.get_end_time()
        other_start_time = interval.get_start_time()
        other_end_time = interval.get_end_time()

        cond = ((self_start_time <= other_start_time and self_end_time >= other_end_time) or (
            self_start_time >= other_start_time and self_end_time <= other_end_time) or (
            self_start_time >= other_start_time and self_end_time >= other_end_time and
            self_start_time <= other_end_time) or (
            self_start_time <= other_start_time and self_end_time >= other_start_time) or (
            self_start_time <= other_end_time and self_end_time >= other_end_time))
        return cond

    cpdef bint spans_in_time(self, double time):
        """Test whether or not `time` is between :attr:`start_time`
        and :attr:`end_time`.

        Parameters
        ----------
        time : float
            The time to test for

        Returns
        -------
        bool:
            Whether or not the time is contained
        """
        return self.get_start_time() <= time <= self.get_end_time()

    def as_arrays(self, dtype=np.float64):
        """Convert this object into a pair of time and intensity arrays, summing
        all peaks at the same time point.

        Returns
        -------
        times : np.ndarray
            The time array for this feature.
        intensities : np.ndarray
            The intensity array for this feature.
        """
        rts = np.array(
            [node.time for node in self.nodes], dtype=dtype)
        signal = np.array([node.total_intensity()
                           for node in self.nodes], dtype=dtype)
        return rts, signal

    def __repr__(self):
        return "%s(%0.4f, %0.2f, %0.2f)" % (
            self.__class__.__name__, self.mz, self.start_time, self.end_time)

    def _copy_chunk(self, nodes, *args, **kwargs):
        x = self.__class__(LCMSFeatureTreeList(nodes))
        x.used_as_adduct = list(self.used_as_adduct)
        return x

    def split_at(self, time):
        """Split this feature at a specific time point into two pieces, before
        and after that point in time.

        Parameters
        ----------
        time : float
            The time point to split at

        Returns
        -------
        before : LCMSFeature
            The component up to the provided time
        after : LCMSFeature
            The component after the provided time
        """
        _, i = self.nodes.find_time(time)
        if self.nodes[i].time < time:
            i += 1
        return LCMSFeature(self.nodes[:i]), LCMSFeature(self.nodes[i:])

    def between_times(self, start, end):
        cdef:
            size_t i, j
            LCMSFeatureTreeNode node
        i = 0
        j = 0
        node = self._find_time(start, &i)
        if node is not None and node.time < start:
            i += 1
        node = self._find_time(end, &j)
        return self._copy_chunk(self.nodes[i:j + 1])

    cpdef list split_sparse(self, double delta_rt=1.):
        """Split this feature at any point where the distance from the
        time between one observed peak and the next exceeds `delta_rt`.

        Parameters
        ----------
        delta_rt : float
            The maximum time between observed peaks to tolerate before
            breaking the feature.

        Returns
        -------
        list of :class:`LCMSFeature`
            The feature components broken at the gaps.
        """
        cdef:
            list chunks, current_chunk
            double last_rt
            LCMSFeatureTreeNode node
            size_t i, n

        chunks = []
        current_chunk = []
        n = self.get_size()
        if n == 0:
            return chunks

        last_rt = self.getitem(0).time
        for i in range(n):
            node = self.getitem(i)
            if (node.time - last_rt) > delta_rt:
                x = self._copy_chunk(current_chunk)

                chunks.append(x)
                current_chunk = []

            last_rt = node.time
            current_chunk.append(node)

        x = self._copy_chunk(current_chunk)

        chunks.append(x)
        for chunk in chunks:
            chunk.created_at = self.created_at

        return chunks

    def truncate_before(self, time):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].time < time:
            i += 1
        self.nodes = LCMSFeatureTreeList(self.nodes[i:])
        self._invalidate()

    def truncate_after(self, time):
        _, i = self.nodes.find_time(time)
        if self.nodes[i].time < time:
            i += 1
        self.nodes = LCMSFeatureTreeList(self.nodes[:i])
        self._invalidate()

    cpdef LCMSFeature clone(self, deep=False, cls=None):
        if cls is None:
            cls = self.__class__
        c = cls(self.nodes.clone(deep=deep), list(self.adducts), list(self.used_as_adduct))
        c.feature_id = self.feature_id
        c.created_at = self.created_at
        return c

    cpdef insert_node(self, LCMSFeatureTreeNode node):
        """Insert a fully formed :class:`LCMSFeatureTreeNode` into this feature
        at the appropriate locaton in the time sequence.

        The node's :attr:`members` attribute will update the peak averager and
        triggers state invalidation.

        Parameters
        ----------
        node : LCMSFeatureTreeNode
            The node to add.
        """
        self._peak_averager._update(node.members)
        self.nodes.insert_node(node)
        self._invalidate()

    cpdef insert(self, PeakBase peak, double time):
        """Insert a peak at a known time into this feature.

        The peak updates the peak averager and triggers state invalidation.

        Parameters
        ----------
        peak : PeakBase
            The a peak with known m/z and intensity
        time : float
            The time point at which this peak was observed
        """
        self._peak_averager.add(peak)
        self.nodes.insert(time, [peak])
        self._invalidate()

    def merge(self, other):
        """Create a copy of this feature and copies all nodes from `other` into it,
        combining information from both features in a new instance.

        Parameters
        ----------
        other : FeatureBase
            Another feature whose nodes (which are distinct from this feature's)
            will be copied into a new copy of this feature.

        Returns
        -------
        LCMSFeature
        """
        new = self.clone()
        for node in other.nodes:
            node = node.clone()
            new.insert_node(node)
        new.created_at = "merge"
        return new

    @property
    def apex_time(self):
        rt, intensity = self.as_arrays()
        return rt[np.argmax(intensity)]

    cpdef bint _eq(self, other):
        if other is None:
            return False
        if len(self) != len(other):
            return False
        else:
            if abs(self.start_time - other.start_time) > 1e-4:
                return False
            elif abs(self.end_time - other.end_time) > 1e-4:
                return False
            else:
                for a, b in zip(self, other):
                    if a != b:
                        return False
                return True

    cpdef bint _ne(self, other):
        return not self._eq(other)

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return self._ne(other)
        else:
            return NotImplemented

    def __reduce__(self):
        return self.__class__, (self.nodes, self.adducts, self.used_as_adduct, self.feature_id)

    def __hash__(self):
        return hash((self.mz, self.start_time, self.end_time))

    def __getitem__(self, i):
        return self.nodes[i]

    cdef LCMSFeatureTreeNode getitem(self, size_t i):
        return self.nodes.getitem(i)

    cpdef double max_intensity(self):
        cdef:
            size_t i, n
            double max_intensity
        max_intensity = 0.0
        n = self.get_size()
        for i in range(n):
            max_intensity = max(self.getitem(i).max_intensity(), max_intensity)
        return max_intensity


cdef class EmptyFeature(FeatureBase):
    def __init__(self, mz):
        FeatureBase.__init__(self, [])
        self._mz = mz
        self._start_time = 0
        self._end_time = INF

    @staticmethod
    cdef EmptyFeature _create(double mz):
        cdef EmptyFeature inst = EmptyFeature.__new__(EmptyFeature)
        inst._mz = mz
        inst.nodes = LCMSFeatureTreeList._create([])
        inst._start_time = 0
        inst._end_time = INF
        return inst

    @property
    def mz(self):
        return self._mz

    cdef double get_mz(self):
        return self._mz

    def __len__(self):
        return 0

    def __getitem__(self, i):
        raise IndexError()

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time


cdef class FeatureSetIterator(object):
    def __init__(self, features, start_time=None, end_time=None):
        cdef:
            size_t n, i

        self.features = list(features)
        n = PyList_GET_SIZE(self.features)
        self.real_features = []
        self.index_list = <size_t*>malloc(sizeof(size_t) * n)
        if self.index_list == NULL:
            raise MemoryError("Failed to allocate index list for LCMS Feature Iterator")
        self.start_time = 0
        self.end_time = INF

        if start_time is not None:
            self.start_time = start_time
        if end_time is not None:
            self.end_time = end_time

        for i in range(n):
            feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
            if feature is None or isinstance(feature, EmptyFeature):
                continue
            self.real_features.append(feature)
            time = feature.get_start_time()
            if (self.start_time < time) and (start_time is None):
                self.start_time = time
            time = feature.get_end_time()
            if (self.end_time > time) and (end_time is None):
                self.end_time = time
            self.index_list[i] = 0

        self.init_indices()
        self.last_time_seen = -1

    cdef int _initialize(self, list features) except 1:
        cdef:
            size_t i, n
        self.features = features
        n = PyList_GET_SIZE(self.features)
        self.real_features = []
        self.index_list = <size_t*>malloc(sizeof(size_t) * n)
        if self.index_list == NULL:
            PyErr_SetString(MemoryError, "Failed to allocate index list for LCMS Feature Iterator")
            return 1
        self.start_time = 0
        self.end_time = INF

        for i in range(n):
            feature = <FeatureBase>PyList_GET_ITEM(self.features, i)
            if feature is None or isinstance(feature, EmptyFeature):
                continue
            self.real_features.append(feature)
            time = feature.get_start_time()
            if (self.start_time < time):
                self.start_time = time
            time = feature.get_end_time()
            if (self.end_time > time):
                self.end_time = time
            self.index_list[i] = 0

        self.init_indices()
        self.last_time_seen = -1
        return 0

    @staticmethod
    cdef FeatureSetIterator _create(list features):
        cdef FeatureSetIterator self
        self = FeatureSetIterator.__new__(FeatureSetIterator)
        self._initialize(features)
        return self

    @staticmethod
    cdef FeatureSetIterator _create_with_threshold(list features, list theoretical_distribution, double detection_threshold):
        cdef:
            FeatureSetIterator self
            TheoreticalPeak p
            size_t i
            bint passed_threshold
            double start_time, end_time
        self = FeatureSetIterator.__new__(FeatureSetIterator)
        self._initialize(features)

        start_time = -INF
        end_time = INF
        for i in range(self.get_size()):
            feature = self.getitem(i)
            if feature is None or isinstance(feature, EmptyFeature):
                continue
            p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)

            passed_threshold = p.intensity > detection_threshold

            time = feature.get_start_time()
            if (start_time < time) and passed_threshold:
                start_time = time
            time = feature.get_end_time()
            if (end_time > time) and passed_threshold:
                end_time = time
        if end_time != INF:
            self.start_time = start_time
            self.end_time = end_time

        self.init_indices()
        self.last_time_seen = -1
        return self

    def __dealloc__(self):
        free(self.index_list)

    cpdef init_indices(self):
        cdef:
            size_t i, ix, n
            FeatureBase f
            LCMSFeatureTreeNode node
        n = self.get_size()
        for i in range(n):
            f = self.getitem(i)
            if f is None:
                self.index_list[i] = 0
                continue
            ix = 0
            if f.get_size() > 0:
                node = f._find_time(self.start_time, &ix)
            self.index_list[i] = ix

    cpdef double get_next_time(self):
        cdef:
            double time, ix_time
            size_t i, n, ix
            FeatureBase f
        time = INF
        n = self.get_size()
        for i in range(n):
            f = self.getitem(i)
            if f is None:
                continue
            ix = self.index_list[i]
            if ix < f.get_size():
                ix_time = f.nodes.getitem(ix).time
                if ix_time < time and ix_time < self.end_time:
                    time = ix_time
        if time == INF:
            return -1
        else:
            return time

    cpdef bint has_more(self):
        cdef:
            size_t j, i, n, ix
            FeatureBase f
            bint done, at_end_time
        j = 0
        n = self.get_size()
        for i in range(n):
            f = self.getitem(i)
            if f is None:
                j += 1
                continue
            ix = self.index_list[i]
            done = ix >= f.get_size()
            if not done:
                at_end_time = f.nodes.getitem(ix).time >= self.end_time
            else:
                at_end_time = done
            if done or at_end_time:
                j += 1
        return j != n

    cdef inline size_t get_size(self):
        return PyList_GET_SIZE(self.features)

    cdef inline FeatureBase getitem(self, size_t i):
        return <FeatureBase>PyList_GET_ITEM(self.features, i)

    def get_index_list(self):
        out = []
        for i in range(self.get_size()):
            out.append(self.index_list[i])
        return out

    cpdef list get_peaks_for_time(self, double time):
        cdef:
            list peaks
            LCMSFeatureTreeNode node
            FeatureBase feature
            size_t i, n, ix
        peaks = []
        i = 0
        n = self.get_size()
        for i in range(n):
            feature = self.getitem(i)
            if feature is not None:
                if feature.get_size() > 0:
                    ix = 0
                    node = feature._find_time(time, &ix)
                    if ix == self.index_list[i] and self.index_list[i] == 0 and node is None:
                        pass
                    elif ix >= self.index_list[i]:
                        self.index_list[i] += 1
                else:
                    if not isinstance(feature, EmptyFeature):
                        raise ValueError("Empty Feature is not of type EmptyFeature")
                    node = None
                if node is not None:
                    peaks.append(
                        node.getpeak(0))
                else:
                    peaks.append(None)
            else:
                peaks.append(None)
        return peaks

    @property
    def current_time(self):
        return self.last_time_seen

    cpdef double get_current_time(self):
        return self.last_time_seen

    cpdef list get_next_value(self):
        cdef:
            double time
            list peaks
        time = self.get_next_time()
        if time < 0:
            return None
        self.last_time_seen = time
        peaks = self.get_peaks_for_time(time)
        return peaks

    def __next__(self):
        cdef:
            list peaks
        peaks = self.get_next_value()
        if peaks is None:
            raise StopIteration()
        return peaks

    def __iter__(self):
        return self

    def __repr__(self):
        return "FeatureSetIterator(%d/%d, %0.3f-%0.3f)" % (
            len(self.real_features), len(self.features), self.start_time, self.end_time)


cdef class RunningWeightedAverage(object):

    cpdef _initialize(self):
        self.current_mean = 0
        self.total_weight = 0
        self.current_count = 0

    def __init__(self, iterable=None):
        self._initialize()
        if iterable is not None:
            self.update(iterable)

    cpdef reset(self):
        self._initialize()

    @staticmethod
    cdef RunningWeightedAverage _create(list peaks):
        cdef:
            RunningWeightedAverage inst

        inst = RunningWeightedAverage.__new__(RunningWeightedAverage)
        inst._initialize()
        if peaks is None:
            return inst
        else:
            inst._update(peaks)
            return inst

    @cython.nonecheck(False)
    @cython.cdivision(True)
    cpdef add(self, PeakBase peak):
        if peak.intensity == 0:
            if self.current_mean == 0 and self.total_weight == 0:
                self.current_mean = peak.mz
                self.total_weight = 1
            else:
                return

        agg = (self.total_weight * self.current_mean) + \
            (peak.mz * peak.intensity)
        self.total_weight += peak.intensity
        self.current_count += 1

        if self.total_weight != 0:
            self.current_mean = agg / (self.total_weight)
        else:
            print("NaN produced in add()")
        return self

    cpdef int feed_from_feature(self, LCMSFeature feature):
        cdef:
            size_t i, j, n, m
            LCMSFeatureTreeNode node
            PeakBase peak

        n = feature.get_size()
        for i in range(n):
            node = feature.getitem(i)
            m = node.get_members_size()
            for j in range(m):
                peak = node.getitem(j)
                self.add(peak)
        return n

    cpdef RunningWeightedAverage update(self, iterable):
        for x in iterable:
            self.add(x)
        return self

    cdef void _update(self, list peaks):
        cdef:
            size_t i, n
            PeakBase p
        n = PyList_GET_SIZE(peaks)
        for i in range(n):
            p = <PeakBase>PyList_GET_ITEM(peaks, i)
            self.add(p)

    def __repr__(self):
        return "%s(%r, %d)" % (self.__class__.__name__, self.current_mean, self.current_count)


cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):

    @staticmethod
    cdef RunningWeightedAverageNeutralMass _create(list peaks):
        cdef:
            RunningWeightedAverageNeutralMass inst

        inst = RunningWeightedAverageNeutralMass.__new__(RunningWeightedAverageNeutralMass)
        inst._initialize()
        if peaks is None:
            return inst
        else:
            inst._update(peaks)
            return inst

    @cython.nonecheck(False)
    @cython.cdivision(True)
    cpdef add(self, PeakBase peak):
        if peak.intensity == 0:
            if self.current_mean == 0 and self.total_weight == 0:
                self.current_mean = (<DeconvolutedPeak?>peak).neutral_mass
                self.total_weight = 1
            else:
                return

        agg = (self.total_weight * self.current_mean) + \
            ((<DeconvolutedPeak?>peak).neutral_mass * peak.intensity)
        self.total_weight += peak.intensity
        self.current_count += 1

        if self.total_weight != 0:
            self.current_mean = agg / (self.total_weight)
        else:
            print("NaN produced in add()")
        return self



import os
import gzip
import pickle
import sys
import logging

try:
    import faulthandler
    faulthandler.enable()
except ImportError:
    pass

from ms_deisotope.config import get_config_dir
from ms_deisotope.task.log_utils import ColoringFormatter
from urllib.request import urlopen

_data_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "test_data"))

_http_data_uri_prefix = "https://raw.githubusercontent.com/mobiusklein/ms_deisotope/master//test_data/"


def _make_logger():
    format_string = '[%(asctime)s] %(levelname).1s | %(name)s | %(message)s'
    colorized_formatter = ColoringFormatter(format_string, datefmt="%H:%M:%S")
    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(colorized_formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    return logger


logger = _make_logger()


def datafile(name: str, force_download: bool = False) -> os.PathLike:
    path = os.path.join(_data_path, name)
    if not os.path.exists(path) or force_download:
        config_dir = get_config_dir()
        test_data_dir = os.path.join(config_dir, "test_data")
        if not os.path.exists(test_data_dir):
            logger.info(
                "Creating user-level test data directory %s", test_data_dir)
            os.makedirs(test_data_dir)
        user_path = os.path.join(test_data_dir, name)
        if not os.path.exists(user_path):
            logger.info(
                "Downloading %s to %s from the remote repository", name, user_path)
            size = 0
            with open(user_path, 'wb') as fh:
                data = urlopen(_http_data_uri_prefix + name).read()
                size = len(data)
                fh.write(data)
            logger.info("Wrote %0.2f MB to %s", size / 1e6, user_path)
        path = user_path
    return path


def gzload(path):
    with gzip.open(path, 'rb') as fh:
        if sys.version_info.major > 2:
            return pickle.load(fh, encoding='latin1')
        else:
            return pickle.load(fh)


def example_scan_bunch():
    import ms_deisotope
    reader = ms_deisotope.MSFileLoader(
        datafile("20150710_3um_AGP_001_29_30.mzML.gz"))
    return reader.next()



# cython: embedsignature=True

cimport cython
from libc.stdlib cimport malloc, free

from ms_peak_picker._c.peak_set cimport PeakSet, FittedPeak, PeakSetIndexed
from ms_peak_picker._c.peak_index cimport PeakIndex

from brainpy._c.isotopic_distribution cimport TheoreticalPeak

from ms_deisotope.constants import (
    ERROR_TOLERANCE as _ERROR_TOLERANCE,
    IGNORE_BELOW as _IGNORE_BELOW,
    TRUNCATE_AFTER as _TRUNCATE_AFTER)
from ms_deisotope._c.scoring cimport IsotopicFitterBase, IsotopicFitRecord
from ms_deisotope._c.averagine cimport (AveragineCache, isotopic_shift, PROTON,
                                        TheoreticalIsotopicPattern, neutral_mass)
from ms_deisotope._c.peak_set cimport DeconvolutedPeak
from ms_deisotope._c.peak_dependency_network.peak_network cimport PeakDependenceGraphBase

from cpython cimport Py_INCREF, PyErr_SetString
from cpython.list cimport PyList_GET_ITEM, PyList_GET_SIZE, PyList_New, PyList_SET_ITEM, PyList_GetSlice
from cpython.tuple cimport PyTuple_GET_ITEM
from cpython.int cimport PyInt_AsLong
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.object cimport PyObject
from cpython.set cimport PySet_Add

import numpy as np
cimport numpy as cnp

cnp.import_array()

import operator


cdef double ERROR_TOLERANCE = _ERROR_TOLERANCE
cdef double IGNORE_BELOW = _IGNORE_BELOW
cdef double TRUNCATE_AFTER = _TRUNCATE_AFTER


cdef size_t count_missed_peaks(list peaklist):
    cdef:
        size_t i
        int n, t
        # FittedPeak peak
        void* peak

    t = n = PyList_GET_SIZE(peaklist)
    for i in range(t):
        # peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        peak = PyList_GET_ITEM(peaklist, i)
        if (<FittedPeak>peak).mz > 1 and (<FittedPeak>peak).intensity > 1:
            n -= 1
    return n


cdef FittedPeak make_placeholder_peak(double mz):
    cdef FittedPeak peak = FittedPeak._create(
        mz, intensity=1.0, signal_to_noise=1.0, peak_count=-1, index=0, full_width_at_half_max=0.0,
        area=1.0, left_width=0.0, right_width=0.0)
    return peak


cdef list drop_placeholders(list peaks):
    """Removes all placeholder peaks from an iterable of peaks

    Parameters
    ----------
    peaks : Iterable of FittedPeak

    Returns
    -------
    list
    """
    cdef:
        size_t i, n
        list retained
        FittedPeak peak
    retained = []
    n = PyList_GET_SIZE(peaks)
    for i in range(n):
        peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
        if peak.mz > 1 and peak.intensity > 1:
            retained.append(peak)
    return retained


cdef FittedPeak first_peak(peaks):
    """Get the first non-placeholder peak in a list of peaks

    Parameters
    ----------
    peaks : Iterable of FittedPeak

    Returns
    -------
    FittedPeak
    """
    cdef:
        size_t i, n
        FittedPeak peak
    n = PyList_GET_SIZE(peaks)
    for i in range(n):
        peak = <FittedPeak>PyList_GET_ITEM(peaks, i)
        if peak.mz > 1 and peak.intensity > 1:
            return peak
    return None


cdef class DeconvoluterBase(object):
    """Base class for all Deconvoluter types. Provides basic configuration for common operations,
    regardless of implementation. Because these methods form the backbone of all deconvolution algorithms,
    this class has a C-extension implementation as well.

    Attributes
    ----------
    peaklist : ms_peak_picker.PeakSet
        The centroided mass spectrum to deconvolute
    scorer : IsotopicFitterBase
        The criterion for evaluating individual isotopic pattern fits
    merge_isobaric_peaks : bool
        If multiple passes produce peaks with identical mass values,
        should those peaks be summed
    minimum_intensity : float
        Experimental peaks whose intensity is below this level will be ignored
        by peak querying methods
    scale_method : str
        The name of the method to use to scale theoretical isotopic pattern intensities
        to match the experimental isotopic pattern. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`.
    use_subtraction : bool
        Whether or not to apply a subtraction procedure to experimental peaks after they
        have been fitted. This is only necessary if the same signal may be examined multiple
        times as in a multi-pass method or when peak dependence is not considered
    verbose : bool
        Produce extra logging information
    """
    def __init__(self, use_subtraction=False, scale_method="sum", merge_isobaric_peaks=True,
                  minimum_intensity=5., *args, **kwargs):
        self.use_subtraction = use_subtraction
        self.scale_method = scale_method
        self.merge_isobaric_peaks = merge_isobaric_peaks
        self.minimum_intensity = minimum_intensity
        self._slice_cache = {}

    cpdef PeakSet between(self, double m1, double m2):
        cdef:
            tuple key
            PyObject* p
            PeakSet region
        key = (m1, m2)
        p = PyDict_GetItem(self._slice_cache, key)
        if p == NULL:
            region = self.peaklist._between(m1, m2)
            PyDict_SetItem(self._slice_cache, key, region)
            return region
        else:
            region = <PeakSet>p
            return region

    cpdef FittedPeak has_peak(self, double mz, double error_tolerance):
        return self._has_peak(mz, error_tolerance)

    @cython.final
    cdef FittedPeak _has_peak(self, double mz, double error_tolerance):
        peak = self.peaklist._has_peak(mz, error_tolerance)
        if peak is None or peak.intensity < self.minimum_intensity:
            return make_placeholder_peak(mz)
        return peak

    cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5):
        cdef:
            list experimental_distribution
            size_t i, n
            double mz
            FittedPeak peak

        n = PyList_GET_SIZE(theoretical_distribution)
        experimental_distribution = PyList_New(n)

        for i in range(n):
            mz = (<TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)).mz
            peak = self._has_peak(mz, error_tolerance)
            Py_INCREF(peak)
            PyList_SET_ITEM(experimental_distribution, i, peak)

        return experimental_distribution

    cpdef scale_theoretical_distribution(self, TheoreticalIsotopicPattern theoretical_distribution,
                                         list experimental_distribution):
        return theoretical_distribution.scale(experimental_distribution, self.scale_method)

    cpdef IsotopicFitRecord _evaluate_theoretical_distribution(self, list experimental,
                                                               TheoreticalIsotopicPattern theoretical,
                                                               FittedPeak peak, int charge):
        """Evaluate a provided theoretical isotopic pattern fit against a
        set of matched experimental peaks.

        Parameters
        ----------
        experimental : list
            A list of experimental fitted peaks to test
        theoretical : :class:`~.TheoreticalIsotopicPattern`
            A list of theoretical isotopic peaks to test
        peak : :class:`~.FittedPeak`
            The seed peak for the isotopic pattern fit
        charge : int
            The target charge state

        Returns
        -------
        :class:`~.IsotopicFitRecord`
        """
        cdef:
            double score
            IsotopicFitRecord fit
        theoretical._scale(experimental, self.scale_method)
        score = self.scorer._evaluate(self.peaklist, experimental, theoretical.peaklist)
        fit = IsotopicFitRecord._create(peak, score, charge, theoretical, experimental, None, 0)
        fit.missed_peaks = count_missed_peaks(fit.experimental)
        return fit

    cpdef list fit_incremental_truncation(self, IsotopicFitRecord seed_fit, double lower_bound):
        """Fit incrementally truncated versions of the seed fit to check to see if a narrower
        theoretical fit matches the data better.

        Parameters
        ----------
        seed_fit : :class:`~.IsotopicFitRecord`
            The original fit to explore truncations of
        lower_bound : float
            The percentage of total signal remaining to stop truncating at.

        Returns
        -------
        :class:`list` of :class:`~.IsotopicFitRecord`
        """
        cdef:
            IsotopicFitRecord fit
            TheoreticalIsotopicPattern pattern
            list fits, patterns, eid
            size_t i, n
            Py_ssize_t k

        patterns = seed_fit.theoretical.incremental_truncation(lower_bound)
        n = PyList_GET_SIZE(patterns)
        fits = PyList_New(n)
        Py_INCREF(seed_fit)
        PyList_SET_ITEM(fits, 0, seed_fit)
        for i in range(1, n):
            pattern = <TheoreticalIsotopicPattern>PyList_GET_ITEM(patterns, i)
            k = pattern.get_size()
            eid = PyList_GetSlice(seed_fit.experimental, 0, k)
            fit = self._evaluate_theoretical_distribution(eid, pattern, seed_fit.seed_peak, seed_fit.charge)
            Py_INCREF(fit)
            PyList_SET_ITEM(fits, i, fit)
        return fits


    cpdef subtraction(self, TheoreticalIsotopicPattern isotopic_cluster, double error_tolerance=2e-5):
        cdef:
            size_t i
            double existing
            TheoreticalPeak peak
            FittedPeak match
        for i in range(isotopic_cluster.get_size()):
            peak = isotopic_cluster.get(i)
            match = self.peaklist._has_peak(peak.mz, error_tolerance)
            if match is not None:
                existing = match.intensity
                match.intensity -= peak.intensity
                if (match.intensity < 0) or (peak.intensity > (existing * 0.7)):
                    match.intensity = 1.

    cpdef bint _check_fit(self, IsotopicFitRecord fit):
        cdef:
            size_t n
        if not has_multiple_real_peaks(fit.experimental) and fit.charge > 1:
            return False
        if self.scorer.reject(fit):
            return False
        return True

    def _merge_peaks(self, peak_list):
        peak_list = sorted(peak_list, key=operator.attrgetter("neutral_mass"))
        if not peak_list:
            return []
        current_peak = peak_list[0]
        merged_peaks = []
        for peak in peak_list[1:]:
            if abs(current_peak.neutral_mass - peak.neutral_mass) < 1e-3 and current_peak.charge == peak.charge:
                current_peak.intensity += peak.intensity
            else:
                merged_peaks.append(current_peak)
                current_peak = peak
        merged_peaks.append(current_peak)
        return merged_peaks

    cdef int _find_next_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):
        cdef:
            double shift, next_peak, prev_peak_mz
            FittedPeak dummy_peak, forward
            size_t i
            size_t start, stop

        shift = isotopic_shift(charge)
        next_peak = mz + (shift * step)
        start = 0
        stop = 0
        self.peaklist._between_bounds(
            next_peak - (next_peak * tolerance),
            next_peak + (next_peak * tolerance),
            &start, &stop)
        for i in range(start, stop):
            forward = self.peaklist.getitem(i)
            prev_peak_mz = forward.mz - (shift * step)
            dummy_peak = make_placeholder_peak(prev_peak_mz)
            PySet_Add(result, (dummy_peak, charge))
        return stop - start

    cdef int _find_previous_putative_peak_inplace(self, double mz, int charge, set result, int step=1, double tolerance=2e-5):
        cdef:
            double shift, prev_peak, prev_peak_mz
            FittedPeak backward
            size_t i, n
            size_t start, stop

        shift = isotopic_shift(charge)
        prev_peak = mz - (shift)
        self.peaklist._between_bounds(
            prev_peak - (prev_peak * tolerance),
            prev_peak + (prev_peak * tolerance),
            &start, &stop)

        for i in range(start, stop):
            backward = self.peaklist.getitem(i)
            prev_peak_mz = backward.mz
            if step == 1:
                self._find_next_putative_peak_inplace(prev_peak_mz, charge, result, 1, tolerance)
            else:
                self._find_previous_putative_peak_inplace(prev_peak_mz, charge, result, step - 1, tolerance)
        return stop - start

    cpdef list _find_next_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):
        """
        Recalibrates the current peak location given the position of the next putative peak
        in a theoretical isotopic cluster.

        Suppose that the peak at `mz` is roughly in the neighborhood of a real isotopic peak,
        but the alignment is bad, so it won't make a good starting point for the search for the
        rest of the peaks in its cluster under a stringent error tolerance.

        However, if we're willing to search for the next putative peak with a more permissive error
        tolerance, which we expect will be properly aligned with the rest of its isotopic cluster,
        we can recalibrate the proper starting peak's mz and use that for isotopic cluster fitting.

        Parameters
        ----------
        mz : float
            Starting m/z value to search from
        charge : int
            Charge state to use when calculating the step size in m/z
        step : int, optional
            The number of steps into the putative isotopic cluster to take. Defaults to 1
        tolerance : float, optional
            The error tolerance to accept for finding supporting peaks.

        Returns
        -------
        list
        """
        cdef:
            double shift, next_peak, prev_peak_mz
            PeakSet peaklist_slice
            list candidates
            FittedPeak dummy_peak, forward
            size_t i, n

        shift = isotopic_shift(charge)
        next_peak = mz + (shift * step)
        peaklist_slice = self.between(
            next_peak - (next_peak * tolerance),
            next_peak + (next_peak * tolerance))
        candidates = []

        n = peaklist_slice.get_size()
        for i in range(n):
            forward = peaklist_slice.getitem(i)
            prev_peak_mz = forward.mz - (shift * step)
            dummy_peak = make_placeholder_peak(prev_peak_mz)
            candidates.append((dummy_peak, charge))
        return candidates

    cpdef list _find_previous_putative_peak(self, double mz, int charge, int step=1, double tolerance=2e-5):
        cdef:
            double shift, prev_peak, prev_peak_mz
            PeakSet peaklist_slice
            list candidates
            FittedPeak backward
            size_t i, n

        shift = isotopic_shift(charge)
        prev_peak = mz - (shift)
        peaklist_slice = self.between(
            prev_peak - (prev_peak * tolerance),
            prev_peak + (prev_peak * tolerance))
        candidates = []
        n = peaklist_slice.get_size()
        for i in range(n):
            backward = peaklist_slice.getitem(i)
            prev_peak_mz = backward.mz
            if step == 1:
                candidates.extend(self._find_next_putative_peak(prev_peak_mz, charge, 1, tolerance))
            else:
                candidates.extend(
                    self._find_previous_putative_peak(prev_peak_mz, charge, step - 1, tolerance))
        return candidates

    def __repr__(self):
        type_name = self.__class__.__name__
        return "%s(peaklist=%s, scorer=%s)" % (type_name, self.peaklist, self.scorer)



cdef bint has_multiple_real_peaks(list peaklist):
    cdef:
        size_t i
        int n
        FittedPeak peak

    n = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        if peak.mz > 1 and peak.intensity > 1:
            n += 1
    return n > 1


cdef class AveragineDeconvoluterBase(DeconvoluterBase):
    """A base class derived from :class:`DeconvoluterBase` which provides some common methods
    for fitting isotopic patterns using an Averagine model.
    """

    def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
                 double minimum_intensity=5., *args, **kwargs):
        super(AveragineDeconvoluterBase, self).__init__(
            use_subtraction, scale_method, merge_isobaric_peaks,
            minimum_intensity, *args, **kwargs)

    cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
                                                         double charge_carrier=PROTON, double truncate_after=0.95,
                                                         double ignore_below=0):
        cdef:
            list eid
            TheoreticalIsotopicPattern tid
            double score
        tid = self.averagine.isotopic_cluster(
            peak.mz, charge, charge_carrier=charge_carrier, truncate_after=truncate_after,
            ignore_below=ignore_below)
        eid = self.match_theoretical_isotopic_distribution(tid.peaklist, error_tolerance=error_tolerance)
        return self._evaluate_theoretical_distribution(eid, tid, peak, charge)

    cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,
                                    double truncate_after=0.95, double ignore_below=0):
        cdef:
            set results
            tuple peak_charge
            IsotopicFitRecord fit
            size_t i
            bint has_incremental_truncation
            double incremental_truncation
            int charge
            list peak_charge_list
        has_incremental_truncation = getattr(self, "incremental_truncation", None) is not None
        if has_incremental_truncation:
            incremental_truncation = self.incremental_truncation
        results = set()
        for obj in peak_charge_set:
            peak_charge = <tuple>obj
            peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
            charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))

            if peak.mz < 1:
                continue

            fit = self.fit_theoretical_distribution(
                     peak, error_tolerance, charge,
                     charge_carrier, truncate_after=truncate_after,
                     ignore_below=ignore_below)
            if not self._check_fit(fit):
                continue
            results.add(fit)
            if has_incremental_truncation:
                results.update(self.fit_incremental_truncation(
                        fit, incremental_truncation))
        return results


cdef class MultiAveragineDeconvoluterBase(DeconvoluterBase):
    """A base class derived from :class:`DeconvoluterBase` which provides some common methods
    for fitting isotopic patterns using multiple Averagine models.
    """
    def __init__(self, bint use_subtraction=False, str scale_method="sum", bint merge_isobaric_peaks=True,
                 double minimum_intensity=5., *args, **kwargs):
        super(MultiAveragineDeconvoluterBase, self).__init__(
            use_subtraction, scale_method, merge_isobaric_peaks,
            minimum_intensity, *args, **kwargs)

    cpdef IsotopicFitRecord fit_theoretical_distribution(self, FittedPeak peak, double error_tolerance, int charge,
                                                         AveragineCache  averagine, double charge_carrier=PROTON,
                                                         double truncate_after=0.95, double ignore_below=0):
        cdef:
            list eid
            TheoreticalIsotopicPattern tid
            double score
        tid = averagine.isotopic_cluster(
            peak.mz, charge, charge_carrier=charge_carrier,
            truncate_after=truncate_after, ignore_below=ignore_below)
        eid = self.match_theoretical_isotopic_distribution(tid.peaklist, error_tolerance=error_tolerance)
        return self._evaluate_theoretical_distribution(eid, tid, peak, charge)

    cpdef set _fit_peaks_at_charges(self, set peak_charge_set, double error_tolerance, double charge_carrier=PROTON,
                                    double truncate_after=0.95, double ignore_below=0):
        cdef:
            list results
            tuple peak_charge
            IsotopicFitRecord fit
            size_t i, j, n_averagine
            int charge
            list peak_charge_list
            bint has_incremental_truncation
            double incremental_truncation

        has_incremental_truncation = getattr(self, "incremental_truncation", None) is not None
        if has_incremental_truncation:
            incremental_truncation = self.incremental_truncation
        results = []
        n_averagine = PyList_GET_SIZE(self.averagines)
        peak_charge_list = list(peak_charge_set)
        for i in range(PyList_GET_SIZE(peak_charge_list)):
            peak_charge = <tuple>PyList_GET_ITEM(peak_charge_list, i)
            peak = <FittedPeak>PyTuple_GET_ITEM(peak_charge, 0)
            charge = PyInt_AsLong(<object>PyTuple_GET_ITEM(peak_charge, 1))

            if peak.mz < 1:
                continue
            for j in range(n_averagine):
                averagine = <AveragineCache>PyList_GET_ITEM(self.averagines, j)
                fit = self.fit_theoretical_distribution(
                    peak, error_tolerance, charge, averagine, charge_carrier,
                    truncate_after=truncate_after, ignore_below=ignore_below)
                # fit.missed_peaks = count_missed_peaks(fit.experimental)
                fit.data = averagine
                if not self._check_fit(fit):
                    continue
                # should we track the best fit for each hypothetical peak charge pair
                # and only add the best one to the result set? This would save time
                # later.
                results.append(fit)
                if has_incremental_truncation:
                    results.extend(
                        self.fit_incremental_truncation(
                            fit, incremental_truncation))

        return set(results)


cdef FittedPeak has_previous_peak_at_charge(DeconvoluterBase peak_collection, FittedPeak peak, int charge, int step, double error_tolerance):
    """Get the `step`th *preceding* peak from `peak` in a isotopic pattern at
    charge state `charge`, or return `None` if it is missing.

    Parameters
    ----------
    peak_collection : DeconvoluterBase
        Peak collection to look up peaks in. Calls :meth:`has_peak` with default accuracy
    peak : ms_peak_picker.FittedPeak
        The peak to use as a point of reference
    charge : int, optional
        The charge state to interpolate from. Defaults to `2`.
    step : int, optional
        The number of peaks along the isotopic pattern to search.

    Returns
    -------
    FittedPeak
    """
    prev = peak.mz - isotopic_shift(charge) * step
    return peak_collection._has_peak(prev, error_tolerance)


cdef FittedPeak has_successor_peak_at_charge(DeconvoluterBase peak_collection, FittedPeak peak, int charge, int step, double error_tolerance):
    """Get the `step`th *succeeding* peak from `peak` in a isotopic pattern at
    charge state `charge`, or return `None` if it is missing.

    Parameters
    ----------
    peak_collection : DeconvoluterBase
        Peak collection to look up peaks in. Calls :meth:`has_peak` with default accuracy
    peak : ms_peak_picker.FittedPeak
        The peak to use as a point of reference
    charge : int, optional
        The charge state to interpolate from. Defaults to `2`.
    step : int, optional
        The number of peaks along the isotopic pattern to search.

    Returns
    -------
    FittedPeak
    """
    nxt = peak.mz + isotopic_shift(charge) * step
    return peak_collection._has_peak(nxt, error_tolerance)


@cython.final
@cython.freelist(10)
cdef class ChargeIterator(object):

    def __init__(self, int lo, int hi):
        self.set_bounds(lo, hi)
        self.make_sequence()

    def __dealloc__(self):
        free(self.values)

    @staticmethod
    cdef ChargeIterator _create(int lo, int hi):
        cdef:
            ChargeIterator inst
        inst = ChargeIterator.__new__(ChargeIterator)
        inst.set_bounds(lo, hi)
        inst.make_sequence()
        return inst

    @staticmethod
    cdef ChargeIterator from_tuple(tuple charge_range):
        cdef:
            ChargeIterator inst
            int a, b

        a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))
        b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
        return ChargeIterator._create(a, b)

    @staticmethod
    cdef ChargeIterator from_quickcharge(tuple charge_range, PeakSet peaks, FittedPeak peak):
        cdef:
            ChargeIterator inst
            int a, b
        a = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 0))
        b = PyInt_AsLong(<object>PyTuple_GET_ITEM(charge_range, 1))
        inst = ChargeIterator.__new__(ChargeIterator)
        inst.set_bounds(a, b)
        inst.sequence_from_quickcharge(peaks, peak)
        return inst

    cpdef sequence_from_quickcharge(self, PeakSet peaks, FittedPeak peak):
        cdef:
            np.ndarray[int, ndim=1] charges
            size_t i, n
        self.release_sequence()
        self.index = 0
        charges = quick_charge(peaks, peak.peak_count, abs(self.lower), abs(self.upper))
        n = charges.shape[0]
        if n == 0:
            self.size = 1
            self.values = <int*>malloc(sizeof(int) * 1)
            if self.values == NULL:
                raise MemoryError("Failed to allocate values array for ChargeIterator.sequence_from_quickcharge")
            self.values[0] = self.sign
        elif charges[0] != 1:
            n += 1
            self.size = n
            self.values = <int*>malloc(sizeof(int) * n)
            if self.values == NULL:
                raise MemoryError("Failed to allocate values array for ChargeIterator.sequence_from_quickcharge")
            self.values[0] = self.sign
            for i in range(1, n):
                self.values[i] = self.sign * charges[i - 1]
        else:
            self.size = n
            self.values = <int*>malloc(sizeof(int) * n)
            if self.values == NULL:
                raise MemoryError("Failed to allocate values array for ChargeIterator.sequence_from_quickcharge")
            for i in range(n):
                self.values[i] = self.sign * charges[i]

    cdef void set_bounds(self, int lo, int hi):
        cdef:
            int abs_lo, abs_hi
        self.sign = -1 if lo < 0 else 1
        abs_lo, abs_hi = abs(lo), abs(hi)
        if abs_lo < abs_hi:
            self.lower = abs_lo
            self.upper = abs_hi
        else:
            self.lower = abs_hi
            self.upper = abs_lo
        self.size = self.upper - self.lower + 1

    cpdef make_sequence(self):
        cdef:
            int v
            size_t i, n
        self.release_sequence()
        n = self.size = self.upper - self.lower + 1
        self.index = 0
        if n == 0:
            return
        self.values = <int*>malloc(sizeof(int) * n)

        for i in range(n):
            self.values[i] = (self.upper - i) * self.sign

    cdef void release_sequence(self):
        if self.values != NULL:
            free(self.values)
            self.values = NULL
        self.size = 0
        self.reset()

    cpdef reset(self):
        self.index = 0

    cpdef bint has_more(self):
        return self.index < self.get_size()

    cpdef int get_next_value(self):
        cdef:
            int value
        value = self.values[self.index]
        self.index += 1
        return value

    cdef size_t get_size(self):
        return self.size

    def __len__(self):
        return self.size

    def __iter__(self):
        return self

    def __next__(self):
        if self.index == self.size:
            raise StopIteration()
        value = self.values[self.index]
        self.index += 1
        return value


@cython.cdivision
cpdef np.ndarray[np.int32_t, ndim=1] quick_charge(FittedPeakCollection peak_set, FittedPeakOrPosition position_spec, int min_charge, int max_charge):
    """An implementation of Hoopman's QuickCharge [1] algorithm for quickly capping charge
    state queries

    Parameters
    ----------
    peak_set : :class:`ms_peak_picker.PeakSet
        The centroided peak set to search
    position_spec : :class:`ms_peak_picker.FittedPeak` or int
        The position of the peak to start the search from. :attr:`ms_peak_picker.FittedPeak.peak_count`
        holds the value that will be used if a peak is passed, otherwise the value is assumed to be an
        index into the `peak_set`.
    min_charge : int
        The minimum charge state to consider
    max_charge : int
        The maximum charge state to consider

    Returns
    -------
    np.ndarray
        The list of feasible charge states

    References
    ----------
    [1] Hoopmann, M. R., Finney, G. L., MacCoss, M. J., Michael R. Hoopmann, Gregory L. Finney,
        and, MacCoss*, M. J., … MacCoss, M. J. (2007). "High-speed data reduction, feature detection
        and MS/MS spectrum quality assessment of shotgun proteomics data sets using high-resolution
        Mass Spectrometry". Analytical Chemistry, 79(15), 5620–5632. https://doi.org/10.1021/ac0700833
    """
    cdef:
        PeakSet peaks
        size_t index, first_index
        int[1000] charges
        np.ndarray[np.int32_t, ndim=1] result
        double min_intensity, diff, raw_charge, remain
        int charge
        int max_charge_found, min_charge_found, result_size
        ssize_t j
        size_t i, n
        bint matched
    if FittedPeakOrPosition is FittedPeak:
        index = position_spec.peak_count
    else:
        index = position_spec
    if FittedPeakCollection is PeakSet:
        peaks = peak_set
    else:
        peaks = peak_set.peaks
    if index > peaks.get_size():
        if peaks.get_size() == 0:
            raise IndexError("peak_set cannot be empty!")
        # Assume we have a slice of the peak list, not the actual complete list, so
        # adjust the index offset from the first peak in the given list.
        first_index = peaks.getitem(0).peak_count
        if (index - first_index) < peaks.get_size() and (index - first_index) >= 0:
            index -= first_index
        else:
            raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, peaks.get_size()))
    n = 1000
    result_size = 0
    # TODO: Make the scaling term a parameterized value
    min_intensity = peaks.getitem(index).intensity / 4.
    for i in range(n):
        charges[i] = 0
    for j in range(index + 1, (peaks.get_size())):
        if peaks.getitem(j).intensity < min_intensity:
            continue
        diff = peaks.getitem(j).mz - peaks.getitem(index).mz
        if diff > 1.1:
            break
        raw_charge = 1 / diff
        charge = <int>(raw_charge + 0.5)
        remain = raw_charge - <int>(raw_charge)
        if 0.2 < remain and remain < 0.8:
            continue
        if (charge < min_charge) or (charge > max_charge):
            continue
        if charges[charge] == 0:
            result_size += 1
        charges[charge] = 1
    if result_size == 0:
        result = np.empty(0, dtype=np.int32)
        return result
    for j in range(index - 1, -1, -1):
        diff = peaks.getitem(index).mz - peaks.getitem(j).mz
        if diff > 1.1:
            break
        raw_charge = 1 / diff
        charge = <int>(raw_charge + 0.5)
        remain = raw_charge - <int>(raw_charge)
        if 0.2 < remain and remain < 0.8:
            continue
        if (charge < min_charge) or (charge > max_charge):
            continue
        if charges[charge] == 0:
            result_size += 1
        charges[charge] = 1
    if result_size <= 0:
        return np.zeros(0, dtype=np.int32)
    result = np.zeros(result_size, dtype=np.int32)

    i = 0
    for j in range(n):
        if charges[j] != 0:
            result[i] = j
            i += 1
    return result


@cython.binding(True)
cpdef set _get_all_peak_charge_pairs(DeconvoluterBase self, FittedPeak peak, double error_tolerance=ERROR_TOLERANCE,
                                     object charge_range=(1, 8), int left_search_limit=3, int right_search_limit=3,
                                     bint recalculate_starting_peak=True, bint use_quick_charge=False):
        """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
        the provided search parameters.

        The search is performed using :func:`has_previous_peak_at_charge`, :func:`has_successor_peak_at_charge`,
        :meth:`_find_previous_putative_peak`, and :meth:`_find_next_putative_peak`.

        Parameters
        ----------
        peak : FittedPeak
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to ERROR_TOLERANCE
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        recalculate_starting_peak : bool, optional
            Whether or not to re-calculate the putative starting peak m/z based upon nearby
            peaks close to where isotopic peaks for `peak` should be. Defaults to True

        Returns
        -------
        set
            The set of all unique candidate (monoisotopic peak, charge state)
        """
        cdef:
            ChargeIterator charge_iterator
            int charge
            size_t i
            set target_peaks
            FittedPeak prev_peak, nxt_peak
            object add_, update_

        if use_quick_charge:
            charge_iterator = ChargeIterator.from_quickcharge(tuple(charge_range), self.peaklist, peak)
        else:
            charge_iterator = ChargeIterator.from_tuple(tuple(charge_range))

        target_peaks = set()
        add_ = target_peaks.add
        update_ = target_peaks.update

        while charge_iterator.has_more():
            charge = charge_iterator.get_next_value()
            add_((peak, charge))

            # Look Left
            for i in range(1, left_search_limit):
                prev_peak = has_previous_peak_at_charge(
                    self, peak, charge, i, error_tolerance)
                if prev_peak is None:
                    continue
                add_((prev_peak, charge))

                if recalculate_starting_peak:
                    # update_(self._find_previous_putative_peak(
                    #     peak.mz, charge, i, 2 * error_tolerance))
                    self._find_previous_putative_peak_inplace(
                        peak.mz, charge, target_peaks, i, 2 * error_tolerance)

            # Look Right
            for i in range(1, right_search_limit):
                nxt_peak = has_successor_peak_at_charge(
                    self, peak, charge, i, error_tolerance)
                if nxt_peak is None:
                    continue
                add_((nxt_peak, charge))


                if recalculate_starting_peak:
                    # update_(self._find_next_putative_peak(
                    #     peak.mz, charge, i, 2 * error_tolerance))
                    self._find_next_putative_peak_inplace(
                        peak.mz, charge, target_peaks, i, 2 * error_tolerance)

            if recalculate_starting_peak:
                for i in range(min(left_search_limit, 2)):
                    # update_(self._find_next_putative_peak(
                    #     peak.mz, charge, step=i, tolerance=2 * error_tolerance))
                    self._find_next_putative_peak_inplace(peak.mz, charge, target_peaks, step=i, tolerance=2 * error_tolerance)

        return target_peaks


@cython.binding(True)
cpdef int _explore_local(DeconvoluterBase self, peak, error_tolerance=ERROR_TOLERANCE,
                         charge_range=(1, 8), left_search_limit=1, right_search_limit=0,
                         charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
                         ignore_below=IGNORE_BELOW):
    """Given a peak, explore the local neighborhood for candidate isotopic fits and add each
    fit above a threshold to the peak dependence graph.

    The threshold assumes that a single peak's neighborhood will contain many, many fits, but
    that only the top `n` scoring fits are worth considering. For now, `n` is fixed at 100 or
    the half number of fits returned, whichever is larger. This is to prevent the fit graph
    from growing out of control and wasting time storing impractical fits. Any fit added to
    the graph will have to pass :attr:`scorer.select` as well, so weak fits will never be added,
    regardless of how many fits are allowed to be inserted.

    Parameters
    ----------
    peak : :class:`~.FittedPeak`
        The peak to start the search from
    error_tolerance : float, optional
        The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
    charge_range : tuple, optional
        The range of charge states to consider. Defaults to (1, 8)
    left_search_limit : int, optional
        The number of steps to search to the left of `peak`. Defaults to 1
    right_search_limit : int, optional
        The number of steps to search to the right of `peak`. Defaults to 0
    charge_carrier : float, optional
        The mass of the charge carrier. Defaults to |PROTON|
    truncate_after : float, optional
        The percent of intensity to ensure is included in a theoretical isotopic pattern
        starting from the monoisotopic peak. This will cause theoretical isotopic patterns
        to be truncated, excluding trailing peaks which do not contribute substantially to
        the overall shape of the isotopic pattern.

    Returns
    -------
    int
        The number of fits added to the graph
    """
    cdef:
        set hold
        set results
        IsotopicFitRecord fit
        size_t n, i, stop
        object fit_saver
        PeakDependenceGraphBase peak_graph

    results = self._fit_all_charge_states(
        peak, error_tolerance=error_tolerance, charge_range=charge_range, left_search_limit=left_search_limit,
        charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)

    peak_graph = <PeakDependenceGraphBase>self.peak_dependency_network

    hold = set()
    for obj in results:
        fit = <IsotopicFitRecord>obj
        if fit.charge > 1 and PyList_GET_SIZE(drop_placeholders(fit.experimental)) == 1:
            continue
        hold.add(fit)

    results = hold
    n = len(results)

    stop = max(min(n // 2, 100), 10)
    if n == 0:
        return 0

    for i in range(stop):
        if len(results) == 0:
            break
        candidate = self.scorer.select.best(results)
        if candidate is None:
                break
        peak_graph.add_fit_dependence(candidate)
        results.discard(candidate)

    return i


@cython.binding(True)
cpdef populate_graph(DeconvoluterBase self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                     left_search_limit=1, right_search_limit=0, charge_carrier=PROTON,
                     truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
    """Visit each experimental peak and execute :meth:`_explore_local` on it with the provided
    parameters, populating the peak dependence graph with all viable candidates.

    Parameters
    ----------
    peak : :class:`~.FittedPeak`
    error_tolerance : float, optional
        The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
    charge_range : tuple, optional
        The range of charge states to consider. Defaults to (1, 8)
    left_search_limit : int, optional
        The number of steps to search to the left of `peak`. Defaults to 1
    right_search_limit : int, optional
        The number of steps to search to the right of `peak`. Defaults to 0
    charge_carrier : float, optional
        The mass of the charge carrier. Defaults to |PROTON|
    truncate_after : float, optional
        The percent of intensity to ensure is included in a theoretical isotopic pattern
        starting from the monoisotopic peak. This will cause theoretical isotopic patterns
        to be truncated, excluding trailing peaks which do not contribute substantially to
        the overall shape of the isotopic pattern.
    """
    cdef:
        size_t i, n
        FittedPeak peak
        dict _priority_map
    _priority_map = self._priority_map
    n = self.peaklist.get_size()
    for i in range(n):
        peak = self.peaklist.getitem(i)
        if peak in _priority_map or peak.intensity < self.minimum_intensity:
            continue
        n = _explore_local(self,
            peak, error_tolerance=error_tolerance, charge_range=charge_range,
            left_search_limit=left_search_limit, right_search_limit=right_search_limit,
            charge_carrier=charge_carrier,
            truncate_after=truncate_after, ignore_below=ignore_below)


@cython.binding(True)
cpdef DeconvolutedPeak _make_deconvoluted_peak(self, IsotopicFitRecord fit, double charge_carrier):
    """Helper method to package a finished :class:`~.IsotopicFitRecord` into a :class:`~.DeconvolutedPeak`

    Arguments
    ---------
    fit: :class:`~.IsotopicFitRecord`
        The fit to package
    charge_carrier: :class:`float`
        The mass of the charge carrier

    Returns
    -------
    :class:`~.DeconvolutedPeak`
    """
    cdef:
        double score, total_abundance, signal_to_noise, full_width_at_half_max
        double most_abundant_mz, most_abundant_mz_intensity, average_mz, area
        double a_to_a2_ratio
        int charge
        list eid, rep_eid, envelope
        TheoreticalIsotopicPattern tid
        FittedPeak fpeak, reference_peak
        size_t i, n

    score = fit.score
    charge = fit.charge
    eid = fit.experimental
    tid = fit.theoretical
    rep_eid = drop_placeholders(eid)

    total_abundance = 0
    signal_to_noise = 0
    full_width_at_half_max = 0
    most_abundant_mz = 0
    most_abundant_mz_intensity = 0
    average_mz = 0
    area = 0
    envelope = []
    n = PyList_GET_SIZE(rep_eid)
    for i in range(n):
        fpeak = <FittedPeak>PyList_GET_ITEM(rep_eid, i)
        if fpeak.intensity > most_abundant_mz_intensity:
            most_abundant_mz_intensity = fpeak.intensity
            most_abundant_mz = fpeak.mz
        total_abundance += fpeak.intensity
        average_mz += fpeak.mz
        area += fpeak.area
        signal_to_noise += fpeak.signal_to_noise
        full_width_at_half_max += fpeak.full_width_at_half_max
    full_width_at_half_max /= n
    signal_to_noise /= n
    average_mz /= n

    n = PyList_GET_SIZE(eid)
    for i in range(n):
        fpeak = <FittedPeak>PyList_GET_ITEM(eid, i)
        envelope.append((fpeak.mz, fpeak.intensity))

    a_to_a2_ratio = 0
    if tid.get_size() > 2:
        a_to_a2_ratio = tid.get(0).intensity / tid.get(2).intensity

    monoisotopic_mass = neutral_mass(
        tid.get_monoisotopic_mz(), charge, charge_carrier)
    reference_peak = first_peak(eid)

    dpeak = DeconvolutedPeak(
        neutral_mass=monoisotopic_mass, intensity=total_abundance, charge=charge,
        signal_to_noise=signal_to_noise,
        index=reference_peak.index if reference_peak is not None else -1,
        full_width_at_half_max=full_width_at_half_max,
        a_to_a2_ratio=a_to_a2_ratio,
        most_abundant_mass=neutral_mass(most_abundant_mz, charge),
        average_mass=neutral_mass(average_mz, charge),
        score=score,
        envelope=envelope,
        mz=tid.monoisotopic_mz, fit=fit,
        area=area)
    return dpeak



# pragma: no cover

from ms_deisotope import mass_charge_ratio

from .text import TextScanSerializerBase


class MS2Serializer(TextScanSerializerBase):
    def __init__(self, stream, sample_name=None, deconvoluted=True):
            super(MS2Serializer, self).__init__(stream, deconvoluted)
            self.sample_name = sample_name
            self.started = False

    def add_global_parameter(self, name, value):
        if self.started:
            raise ValueError("Cannot add global parameter if scan data has begun being written")
        self.stream.write("H\t{0}\t{1}\n".format(name, value))

    def _format_parameter(self, key, value, symbol):
        return "{2}\t{0}\t{1}\n".format(str(key).upper(), str(value), symbol)

    def _add_parameter(self, name, value, symbol):
        self.stream.write(self._format_parameter(name, value, symbol))

    def add_parameter(self, name, value, symbol="I"):
        self._add_parameter(name, value, symbol)

    def save_scan_bunch(self, bunch, **kwargs):
        for scan in bunch.products:
            self.write_scan(*self.prepare_scan_data(scan))

    def format_peak_vectors(self, scan):
        if self.deconvoluted:
            mz = [i.mz for p in scan.deconvoluted_peak_set for i in p.envelope]
            intensity = [i.intensity for p in scan.deconvoluted_peak_set for i in p.envelope]
            return (mz, intensity, None)
        else:
            mz = [p.mz for p in scan.peak_set]
            intensity = [p.intensity for p in scan.peak_set]
            return mz, intensity, None

    def write_header(self, header_dict):
        mass = header_dict['precursor_neutral_mass']
        charge = header_dict['precursor_charge']
        mz = mass_charge_ratio(mass, charge)
        second_index = header_dict.scan.index
        try:
            first_index = header_dict.scan.precursor_information.precursor.index
        except AttributeError:
            first_index = second_index
        self.stream.write("S\t%r\t%r\t%r\n" % (first_index, second_index, mz))
        self.stream.write("Z\t%r\t%r\n" % (charge, mass))



import threading
import logging

from .scan.scan import Scan
from .scan.loader import RandomAccessScanSource
from .infer_type import MSFileLoader
from ._compression import get_opener


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PerThreadFileHandle(object):
    def __init__(self, path, mode='rb', encoding=None, opener=None):
        if opener is None:
            opener = get_opener
        self.local = threading.local()
        self._path = path
        self._mode = mode
        self._encoding = encoding
        self._opener = opener

    @property
    def file(self):
        try:
            return self.local.file
        except AttributeError:
            return self._open()

    def __len__(self):
        return len(self.file)

    def __iter__(self):
        return iter(self.file)

    def __getitem__(self, i):
        return self.file[i]

    def __reduce__(self):
        return self.__class__, (self._path, self._mode, self._encoding, self._opener)

    def _open(self):
        self.local.file = self._opener(self._path)
        logger.debug("Opening %r on Thread %r", self._path,
                     threading.current_thread())
        return self.local.file

    def __getattr__(self, attrib):
        if attrib in ('local', '_path', '_mode', '_encoding', '_opener'):
            raise AttributeError(attrib)
        return object.__getattribute__(self.file, attrib)



class ThreadsafeScanSource(object):

    def __init__(self, source_file, opener=None):
        if opener is None:
            opener = MSFileLoader
        self._source_file_name = source_file
        self._opener = opener
        self._threadlocal_store = threading.local()

    def _reader(self):
        try:
            return self._threadlocal_store.source
        except AttributeError:
            self._threadlocal_store.source = self._patch(self._opener(self._source_file_name))
            return self._threadlocal_store.source

    def _make_scan(self, data):
        return Scan(data, self).bind(self)

    def _patch(self, reader):
        """Replace the _make_scan method of `reader`
        """
        reader._make_scan = self._make_scan
        return reader

    def __getattr__(self, name):
        reader = self._reader()
        value = getattr(reader, name)
        return value

    def __reduce__(self):
        return self.__class__, (self._source_file_name, self._opener)



#import copy
import ctypes
from ctypes import*
#from array import *
from .MassLynxRawReader import *

class MassLynxParameters(object):
    """Create parameter"""

    def __init__(self):
        self._mlParameters = c_void_p()              # instance variable
        self._codeHandler = MassLynxCodeHandler()
        self._stringHandler = MassLynxStringHandler()
        createParameters = MassLynxRawReader.massLynxDll.createParameters
        createParameters.argtypes = [POINTER(c_void_p)]
        self.CheckReturnCode(createParameters(self._mlParameters))

    # destroy the samplelist
    def __del__(self):
        destroyParameters = MassLynxRawReader.massLynxDll.destroyParameters
        destroyParameters.argtypes = [c_void_p]
        destroyParameters( self._mlParameters )

    # get the parameter object
    def GetParameters(self):
        return self._mlParameters

    # handle the retun code - no exceptions
    def CheckReturnCode( self, code ):
        return self._codeHandler.CheckReturnCode(code, False)

    # Set the key value pair
    def Set( self, key, value ):
        # is it a bool
        if isinstance(value, (bool)):
            if ( value ):
                value = 1
            else:
                value = 0

        # convert to string and encode
        bytes = str.encode(str(value))

        setParameterValue = MassLynxRawReader.massLynxDll.setParameterValue
        setParameterValue.argtypes = [c_void_p, c_int, c_char_p]
        setParameterValue( self._mlParameters, key, bytes )

        return self

    def Get( self, key ):
        # call the dll
        value = c_char_p()
        getParameterValue = MassLynxRawReader.massLynxDll.getParameterValue
        getParameterValue.argtypes = [c_void_p, c_int, POINTER(c_char_p)]
        getParameterValue(self._mlParameters, key, value )

        return self._stringHandler.ToString(value, False)